from pathlib import Path
from typing import Any, Union

from joblib import dump, load

from mlpype.base.serialiser.serialiser import Serialiser


class JoblibSerialiser(Serialiser):
    def serialise(self, object: Any, file: Union[str, Path]) -> None:
        """Serialise the given object to the given file.

        Args:
            object (Any): The object to serialise.
            file (Union[str, Path]): The file to serialise to.
        """
        dump(object, file)

    def deserialise(self, file: Union[str, Path]) -> Any:
        """Deserialise the object in the given file.

        Args:
            file (Union[str, Path]): The file to deserialise.

        Returns:
            Any: The python object stored in the file.
        """
        return load(file)



from abc import ABC, abstractmethod
from typing import Generic, Tuple, TypeVar

Data = TypeVar("Data")


class Operator(ABC, Generic[Data]):
    def __init__(self) -> None:
        """An Operator that can been fitted to Data and transform new data.

        This is not meant to be extended, but should work well enough to help
        people identify what their Pipes need as input.
        """
        super().__init__()

    @abstractmethod
    def fit(self, *data: Data) -> "Operator":
        """Fit the Operator to the given Data.

        Returns:
            Operator: this object.
        """
        raise NotImplementedError

    @abstractmethod
    def transform(self, *data: Data) -> Tuple[Data, ...]:
        """Transforms the data using this Operator.

        You do not need to return the same amount of Data as you
        used for input.

        Returns:
            Tuple[Data]: A Tuple of Data elements, the transformed
                Data.
        """
        raise NotImplementedError

    def inverse_transform(self, *data: Data) -> Tuple[Data, ...]:
        """Optionally, you can include an option to inverse transform the data.

        Returns:
            Tuple[Data]: An Tuple of Data elements, the inverse transformed
                Data.
        """
        return data



from pathlib import Path

from fastapi.applications import FastAPI

from mlpype.fastapi.deploy import mlpypeApp


def load_app() -> FastAPI:
    """Wrapper function to create a FastAPI server from a fixed location in this package.

    Based on mlpypeApp.

    If you create a python file F where this App is loaded:
    ```
    from <model name> import load_app
    app = load_app()
    ```

    You can host this model using a command like:
    uvicorn <python package path to file F>:app

    Returns:
        FastAPI: The FastAPI/mlpypeApp server used to host the model.
    """
    folder = Path(__file__).parent.parent / "outputs"
    return mlpypeApp("model", folder=folder).create_app()



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

Data = TypeVar("Data")


class DataSink(ABC, Generic[Data]):
    @abstractmethod
    def write(self, data: Data) -> None:
        """Writes data to a given source."""
        raise NotImplementedError



from pathlib import Path

from mlpype.base.deploy.wheel import WheelExtension

FastApiExtension = WheelExtension(
    "fastapi", [(Path(__file__).parent / "wheel" / "helpers.py", ["load_app"])], ["mlpype.fastapi"]
)



from pathlib import Path

from mlpype.base.deploy import Inferencer


def load_model() -> Inferencer:
    """Wrapper function to load an Inferencer from a fixed location in a package.

    Returns:
        Inferencer: The Inferencer contained in this package.
    """
    folder = Path(__file__).parent.parent / "outputs"
    print(f"Loading inferencer from {str(folder)}")
    return Inferencer.from_folder(folder)



import json
import os
import subprocess
import sys
from argparse import ArgumentParser
from logging import getLogger
from pathlib import Path
from typing import Any, Dict, List, Optional, Type, Union

from mlpype.base.constants import Constants
from mlpype.base.data import DataCatalog, DataSet
from mlpype.base.evaluate import BaseEvaluator
from mlpype.base.experiment.argument_parsing import add_args_to_parser_for_pipeline
from mlpype.base.logger import ExperimentLogger
from mlpype.base.model import Model
from mlpype.base.pipeline import Pipeline
from mlpype.base.pipeline.type_checker import TypeCheckerPipe
from mlpype.base.serialiser import JoblibSerialiser, Serialiser
from mlpype.base.utils.parsing import get_args_for_prefix


class Experiment:
    def __init__(
        self,
        data_sources: Dict[str, Union[DataCatalog, DataSet]],
        model: Model,
        pipeline: Pipeline,
        evaluator: BaseEvaluator,
        logger: ExperimentLogger,
        input_type_checker: TypeCheckerPipe,
        output_type_checker: TypeCheckerPipe,
        serialiser: Optional[Serialiser] = None,
        output_folder: Union[Path, str] = "outputs",
        additional_files_to_store: Optional[List[Union[str, Path]]] = None,
        parameters: Optional[Dict[str, Any]] = None,
    ):
        """The core of the mlpype library: run a standardised ML experiment with the given parameters.

        It is highly recommended to use one of the class methods to initialise this object:
            - from_dictionary: If you've already got your parameters in the right form:
                - `model__<arg>` for model parameters
                - `pipeline__<pipe_name>__<arg> for pipeline parameters.

        Args:
            data_sources (Dict[str, Dict[str, Union[DataCatalog, DataSet]]]): The
                DataCatalog or DataSets to use, in DataSource form. Should contain at least a 'train' DataSet.
                DataCatalogs will be initialised (`read`) in the beginning of the run. It is recommended to use
                DataCatalog in distributed cases, but for quick experimentation DataSets tend to be more useful.
                Please note that the DataSet will be changed by the Pipeline and Model. If you want to use a
                DataSet, you should make sure you don't modify the original data by reference or by key.
            model (Model): The Model to fit.
            pipeline (Pipeline): The Pipeline to use to transform data before feeding it to the Model.
            evaluator (BaseEvaluator): The evaluator to test how good your Model performs.
            logger (ExperimentLogger): The experiment logger to make sure you record how well your experiment worked,
                and log any artifacts such as the trained model.
            serialiser (Optional[Serialiser]): The serialiser to serialise any Python objects (expect the Model).
                Defaults to a joblib serialiser.
            output_folder: (Union[Path, str]): The output folder to log artifacts to. Defaults to "outputs".
            input_type_checker: (TypeCheckerPipe): A type checker for all input data. Will be used to verify incoming
                data and standardise the order of data. Will be used later to help serialise/deserialise data. Only
                select datasets required to do a run during inference.
            output_type_checker: (TypeCheckerPipe): A type checker for all output data. Will be used to verify outgoing
                data and standardise the order of data. Will be used later to help serialise/deserialise data.
            additional_files_to_store (Optional[List[Union[str, Path]]]): Extra files to store, such as python files.
                It is possible to select a directory as well, not just individual files.
                Defaults to no extra files (None).
            parameters (Optional[Dict[str, Any]]): Any parameters to log as part of this experiment.
                Defaults to None.
        """
        assert "train" in data_sources, "Must provide a 'train' entry in the data_sources dictionary."
        self.logger = getLogger(__name__)
        if serialiser is None:
            serialiser = JoblibSerialiser()

        if additional_files_to_store is None:
            additional_files_to_store = []
        if parameters is None:
            parameters = {}
            self.logger.warning(
                """It is highly recommended to provide the parameters used to initialise your
run here for logging purposes. Consider using the `from_command_line` or
`from_dictionary` initialisation methods"""
            )
        if isinstance(output_folder, str):
            output_folder = Path(output_folder)

        self.data_sources = data_sources
        self.model = model
        self.pipeline = pipeline
        self.evaluator = evaluator
        self.input_type_checker = input_type_checker
        self.output_type_checker = output_type_checker

        self.experiment_logger = logger
        self.parameters = parameters
        self.serialiser = serialiser
        self.output_folder = output_folder
        self.additional_files_to_store = additional_files_to_store

    def run(self) -> Dict[str, Dict[str, Union[str, float, int, bool]]]:
        """Execute the experiment.

        Returns:
            Dict[str, Union[str, float, int, bool]]: The performance metrics of this run.
        """
        with self.experiment_logger:
            self.logger.info("Load data")
            datasets = {
                name: data_source_set.read() if isinstance(data_source_set, DataCatalog) else data_source_set
                for name, data_source_set in self.data_sources.items()
            }

            self.logger.info("Create input type checker")
            self.input_type_checker.fit(datasets["train"])
            for ds in datasets.values():
                self.input_type_checker.transform(ds)

            self.logger.info("Fit pipeline")
            self.pipeline.fit(datasets["train"])

            self.logger.info("Transform data")
            transformed = {name: self.pipeline.transform(data, is_inference=False) for name, data in datasets.items()}

            self.logger.info("Fit model")
            self.model.fit(transformed["train"])

            self.logger.info("Evaluate model")
            metrics = {name: self.evaluator.evaluate(self.model, data) for name, data in transformed.items()}

            self.logger.info("Create output type checker")
            predicted_train = self.model.transform(transformed["train"])
            self.output_type_checker.fit(predicted_train)

            self.logger.info("Log results: metrics, parameters")
            for dataset_name, metric_set in metrics.items():
                self.experiment_logger.log_metrics(dataset_name, metric_set)
            self.experiment_logger.log_parameters(self.parameters)

            self.logger.info("Log results: pipeline, model, serialiser")
            self._log_run()
        return metrics

    def _log_run(self) -> None:
        self._create_output_folders()

        of = self.output_folder
        self.experiment_logger.log_model(self.model, of / Constants.MODEL_FOLDER)
        self.experiment_logger.log_artifact(of / Constants.PIPELINE_FILE, self.serialiser, object=self.pipeline)
        self.experiment_logger.log_artifact(
            of / Constants.INPUT_TYPE_CHECKER_FILE, self.serialiser, object=self.input_type_checker
        )
        self.experiment_logger.log_artifact(
            of / Constants.OUTPUT_TYPE_CHECKER_FILE, self.serialiser, object=self.output_type_checker
        )

        # log requirements.txt
        self._log_requirements()

        # extra py files
        self._log_extra_files()
        self.logger.info("Done")

        # log serialiser using a JoblibSerialiser
        jl_serialiser = JoblibSerialiser()
        self.experiment_logger.log_artifact(of / Constants.SERIALISER_FILE, jl_serialiser, object=self.serialiser)

    def _log_requirements(self) -> None:
        python_version = f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
        version_info = {
            "python_version": python_version,
            "major": sys.version_info.major,
            "minor": sys.version_info.minor,
            "micro": sys.version_info.micro,
        }
        version_file = self.output_folder / Constants.PYTHON_VERSION_FILE
        with open(version_file, "w") as f:
            json.dump(version_info, f)
        self.experiment_logger.log_file(version_file)

        reqs = subprocess.check_output([sys.executable, "-m", "pip", "freeze"]).decode()
        requirements_file = self.output_folder / Constants.REQUIREMENTS_FILE
        with open(requirements_file, "w") as f:
            f.write(reqs)
        self.experiment_logger.log_file(requirements_file)

    def _log_extra_files(self) -> None:
        """Logs the extra files for an experiment, as specified in the constructor."""
        paths_to_log = []
        self.logger.info("Log extra files")
        for extra_file in self.additional_files_to_store:
            relative_path = Path(extra_file).relative_to(os.getcwd())
            self.experiment_logger.log_local_file(relative_path, self.output_folder / relative_path)
            paths_to_log.append(str(relative_path))

        self.logger.info("Log `extra files`-file")
        extra_files_file = self.output_folder / Constants.EXTRA_FILES
        with open(extra_files_file, "w") as f:
            data = {"paths": paths_to_log}
            json.dump(data, f)
        # Make sure the file path is closed before logging anything, to make sure all writes have flushed.
        self.experiment_logger.log_file(extra_files_file)

    def _create_output_folders(self) -> None:
        of = self.output_folder
        of.mkdir(exist_ok=True, parents=True)
        (of / Constants.MODEL_FOLDER).mkdir(exist_ok=True)

    def copy(self, parameters: Dict[str, Any], seed: int = 1) -> "Experiment":
        """Create a fresh copy of this Experiment.

        The Model & Pipeline will be recreated, so any trained versions will be
        re-initialised.

        Old parameters will be kept, but overwritten if new values are provided.
        This makes it easier to make templates for hyperparameter tuning.

        Args:
            parameters (Dict[str, Any]): New parameters for the Model and Pipeline
                to be re-initialised with.
            seed (int, optional): Training seed. Defaults to 1.

        Returns:
            Experiment: A copy of this experiment, intialised with the new set
                of parameters.
        """
        new_params = self.parameters.copy()
        new_params.update(parameters)
        model_class = self.model.__class__

        model_args = get_args_for_prefix("model__", new_params)
        model = model_class(seed=seed, inputs=self.model.inputs, outputs=self.model.outputs, **model_args)

        pipeline_args = get_args_for_prefix("pipeline__", new_params)
        new_pipeline = self.pipeline.copy(pipeline_args)

        return Experiment(
            # We might want to also deep copy data_sources, evaluator, type checkers, and experiment_logger.
            data_sources=self.data_sources,
            model=model,
            pipeline=new_pipeline,
            evaluator=self.evaluator,
            logger=self.experiment_logger,
            serialiser=self.serialiser,
            output_folder=self.output_folder,
            input_type_checker=self.input_type_checker,
            output_type_checker=self.output_type_checker,
            additional_files_to_store=self.additional_files_to_store,
            parameters=new_params,
        )

    @classmethod
    def from_dictionary(
        cls,
        data_sources: Dict[str, DataCatalog],
        model_class: Type[Model],
        pipeline: Pipeline,
        evaluator: BaseEvaluator,
        logger: ExperimentLogger,
        serialiser: Serialiser,
        input_type_checker: TypeCheckerPipe,
        output_type_checker: TypeCheckerPipe,
        model_inputs: List[str],
        model_outputs: List[str],
        parameters: Dict[str, Any],
        output_folder: Union[Path, str] = "outputs",
        seed: int = 1,
        additional_files_to_store: Optional[List[Union[str, Path]]] = None,
    ) -> "Experiment":
        """Creates an Experiment from a dictionary with parameters.

        Args:
            data_sources (Dict[str, DataCatalog]): The DataSets to use, in DataSource form.
                Should contain at least a 'train' DataSet. These will be initialised in the beginning of the run.
            model_class (Type[Model]): The class of the Model to fit.
            pipeline (Pipeline): The Pipeline to use to transform data before feeding it to the Model.
            evaluator (BaseEvaluator): The evaluator to test how good your Model performs.
            logger (ExperimentLogger): The experiment logger to make sure you record how well your experiment worked,
                and log any artifacts such as the trained model.
            serialiser (Serialiser): The serialiser to serialise any Python objects (expect the Model).
            output_folder: (Union[Path, str]): The output folder to log artifacts to.
            input_type_checker: (TypeCheckerPipe): A type checker for all input data. Will be used to verify incoming
                data and standardise the order of data. Will be used later to help serialise/deserialise data.
            output_type_checker: (TypeCheckerPipe): A type checker for all output data. Will be used to verify outgoing
                data and standardise the order of data. Will be used later to help serialise/deserialise data.
            model_inputs: (List[str]): Input dataset names to the model.
            model_outputs: (List[str]): Output dataset names to the model.
            seed (int): The RNG seed to ensure reproducability.
            parameters (Optional[Dict[str, Any]]): Any parameters to log as part of this experiment.
                Defaults to None.
            additional_files_to_store (Optional[List[Union[str, Path]]]): Extra files to store, such as python files.
                Defaults to no extra files (None).

        Returns:
            Experiment: An Experiment created with the given parameters.
        """
        model_args = get_args_for_prefix("model__", parameters)
        model = model_class(seed=seed, inputs=model_inputs, outputs=model_outputs, **model_args)

        pipeline_args = get_args_for_prefix("pipeline__", parameters)
        pipeline.reinitialise(pipeline_args)

        return Experiment(
            data_sources=data_sources,
            model=model,
            pipeline=pipeline,
            evaluator=evaluator,
            logger=logger,
            serialiser=serialiser,
            output_folder=output_folder,
            input_type_checker=input_type_checker,
            output_type_checker=output_type_checker,
            additional_files_to_store=additional_files_to_store,
            parameters=parameters,
        )

    @classmethod
    def from_command_line(
        cls,
        data_sources: Dict[str, DataCatalog],
        model_class: Type[Model],
        pipeline: Pipeline,
        evaluator: BaseEvaluator,
        logger: ExperimentLogger,
        serialiser: Serialiser,
        input_type_checker: TypeCheckerPipe,
        output_type_checker: TypeCheckerPipe,
        model_inputs: List[str],
        model_outputs: List[str],
        output_folder: Union[Path, str] = "outputs",
        seed: int = 1,
        additional_files_to_store: Optional[List[Union[str, Path]]] = None,
        fixed_arguments: Optional[Dict[str, Any]] = None,
    ) -> "Experiment":
        """Automatically initialises an Experiment from command line arguments.

        Args:
            data_sources (Dict[str, DataCatalog]): The DataSets to use, in DataSource form.
                Should contain at least a 'train' DataSet. These will be initialised in the beginning of the run.
            model_class (Type[Model]): The class of the Model to fit.
            pipeline (Pipeline): The Pipeline to use to transform data before feeding it to the Model.
            evaluator (BaseEvaluator): The evaluator to test how good your Model performs.
            logger (ExperimentLogger): The experiment logger to make sure you record how well your experiment worked,
                and log any artifacts such as the trained model.
            serialiser (Serialiser): The serialiser to serialise any Python objects (expect the Model).
            output_folder: (Union[Path, str]): The output folder to log artifacts to.
            input_type_checker: (TypeCheckerPipe): A type checker for all input data. Will be used to verify incoming
                data and standardise the order of data. Will be used later to help serialise/deserialise data.
            output_type_checker: (TypeCheckerPipe): A type checker for all output data. Will be used to verify outgoing
                data and standardise the order of data. Will be used later to help serialise/deserialise data.
            model_inputs: (List[str]): Input dataset names to the model.
            model_outputs: (List[str]): Output dataset names to the model.
            seed (int): The RNG seed to ensure reproducability.
            additional_files_to_store (Optional[List[Union[str, Path]]], optional): Extra files to store,
                such as python files. Defaults to no extra files (None).
            fixed_arguments (Optional[Dict[str, Any]]): Arguments that won't be read from command line.
                Useful to pass complex objects that you don't want to optimize. Think of classes, loss functions,
                metrics, etc. Any value in this dict will overwrite values from the command line. These need to be
                presented in the same `model__<arg>` or `pipeline__<step>__<arg>` format as the command line arguments.


        Returns:
            Experiment: An Experiment created with the given parameters from the command line.
        """
        arg_parser = cls._get_cmd_args(model_class, pipeline)
        parsed_args, _ = arg_parser.parse_known_args()

        arguments = parsed_args.__dict__
        if fixed_arguments is not None:
            arguments.update(fixed_arguments)

        return cls.from_dictionary(
            data_sources=data_sources,
            model_class=model_class,
            pipeline=pipeline,
            evaluator=evaluator,
            logger=logger,
            serialiser=serialiser,
            output_folder=output_folder,
            input_type_checker=input_type_checker,
            output_type_checker=output_type_checker,
            additional_files_to_store=additional_files_to_store,
            parameters=parsed_args.__dict__,
            seed=seed,
            model_inputs=model_inputs,
            model_outputs=model_outputs,
        )

    @classmethod
    def _get_cmd_args(cls, model_class: Type[Model], pipeline: Pipeline) -> ArgumentParser:
        parser = ArgumentParser()

        model_class.get_parameters(parser)
        add_args_to_parser_for_pipeline(parser, pipeline)
        return parser

    def __str__(self) -> str:
        """Create string representation of this Experiment.

        Returns:
            str: A string representation of this Experiment.
        """
        data_part = "datasets:\n" + "\n\t".join(
            [
                f"""\t{name}:
{ds_source.__str__(indents=2)}"""
                for name, ds_source in self.data_sources.items()
            ]
        )

        pipeline_part = "pipeline:\n" + self.pipeline.__str__(indents=1)
        model_part = f"model:\n\t{str(self.model)}"
        logger_part = f"logger:\n\t{str(self.experiment_logger)}"
        evaluator_part = f"evaluator:\n{self.evaluator.__str__(indents=1)}"

        return "\n\n".join([data_part, pipeline_part, model_part, logger_part, evaluator_part])



import os
from pathlib import Path
from typing import List, Union

from setuptools import setup


def package_files(directory: Union[Path, str]) -> List[str]:
    """Find all files from the output directory of an Experiment run to include.

    Args:
        directory (Union[Path, str]): The main output directory to search.

    Returns:
        List[str]: The files to include.
    """
    paths = []
    for (path, _, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join("..", path, filename))
    return paths


extra_files = package_files(Path().parent.absolute())

setup(
    name="{package_name}",
    install_requires=["{install_requires}"],
    version="{version}",
    include_package_data=True,
    package_data={"": extra_files},
)



from typing import Any

from keras import Model
from keras.layers import Dense
from tensorflow import Tensor  # type: ignore


class MLPKeras(Model):
    def __init__(
        self,
        output_size: int,
        n_layers: int,
        layer_size: int,
        activation: str = None,
        output_activation: str = None,
        *args: Any,
        **kwargs: Any,
    ):
        """A simple model showing how to use a MLP from Keras with mlpype.

        Args:
            output_size (int): The dimensions of the output
            n_layers (int): The number of hidden layers
            layer_size (int): The dimensions of hidden layers.
            activation (str, optional): The activation function of hidden layers. Defaults to None.
            output_activation (str, optional): The output activation function. Defaults to None.
        """
        super().__init__(*args, **kwargs)
        assert n_layers >= 0, "Need at least 1 layer in the model."

        activations = [activation for _ in range(n_layers)] + [output_activation]
        sizes = [layer_size for _ in range(n_layers)] + [output_size]
        self._layers = [Dense(size, activation=act) for act, size in zip(activations, sizes)]

    def call(self, inputs: Tensor) -> Tensor:
        """Standard Keras model API call."""
        for layer in self._layers:
            inputs = layer(inputs)
        return inputs



from typing import Any, Dict

from .keras_pype_model import KerasPypeModel
from .mlp_keras import MLPKeras


class MLPPypeModel(KerasPypeModel[MLPKeras]):
    """A Keras model integrated with mlpype's APIs."""

    def _init_model(self, args: Dict[str, Any]) -> MLPKeras:
        return MLPKeras(**args)



from typing import Any, Dict

from sklearn.linear_model import LinearRegression

from mlpype.sklearn.model.sklearn_model import SklearnModel


class LinearRegressionModel(SklearnModel[LinearRegression]):
    def _init_model(self, args: Dict[str, Any]) -> LinearRegression:
        return LinearRegression(**args)



from pathlib import Path
from typing import Any, Dict, List, Type

from xgboost.sklearn import XGBClassifier

from mlpype.sklearn.model import SklearnModel


class XGBClassifierModel(SklearnModel[XGBClassifier]):
    XGB_MODEL_FILE = "model.txt"

    def _init_model(self, args: Dict[str, Any]) -> XGBClassifier:
        return XGBClassifier(**args)

    def _save(self, folder: Path) -> None:
        assert isinstance(self.model, XGBClassifier)
        self.model.save_model(folder / self.XGB_MODEL_FILE)

    @classmethod
    def _load(cls: Type["XGBClassifierModel"], folder: Path, inputs: List[str], outputs: List[str]) -> "SklearnModel":
        model = XGBClassifier()
        model.load_model(folder / cls.XGB_MODEL_FILE)
        return cls(inputs=inputs, outputs=outputs, model=model, seed=1)



from abc import ABC, abstractmethod
from typing import Dict, Generic, Optional, TypeVar, Union

from mlpype.base.data import DataSet
from mlpype.base.model import Model
from mlpype.base.pipeline import Pipeline

Data = TypeVar("Data")


class BaseEvaluator(Generic[Data], ABC):
    def __init__(
        self,
    ) -> None:
        """Evaluates a Model on the given data."""

    @abstractmethod
    def evaluate(
        self, model: Model, data: DataSet, pipeline: Optional[Pipeline] = None
    ) -> Dict[str, Union[str, float, int, str, bool]]:
        """Evaluate the given model on the given dataset.

        We assume the model does not need to be transformed anymore if pipeline is None.

        Args:
            model (Model): The Model to evaluate.
            data (DataSet): The Dataset to use to evaluate the model.
            pipeline (Optional[Pipeline]): If not None, this will be used to transform `data` first.

        Returns:
            Dict[str, Union[str, float, int, str, bool]]: A dictionary of metric_name-value pairs. The result
                of the evaluation.
        """

    @abstractmethod
    def __str__(self, indents: int = 0) -> str:
        """Create string representation of this Evaluator.

        Args:
            indents (int, optional): The number of preceding tabs. Defaults to 0.

        Returns:
            str: A string representation of this Evaluator.
        """



from typing import Any, Dict

from sklearn.linear_model import LogisticRegression

from mlpype.sklearn.model.sklearn_model import SklearnModel


class LogisticRegressionModel(SklearnModel[LogisticRegression]):
    def _init_model(self, args: Dict[str, Any]) -> LogisticRegression:
        return LogisticRegression(**args)



from tensorflow import Tensor  # type: ignore

from mlpype.base.data.data_source import DataSource


class TensorSource(DataSource[Tensor]):
    def __init__(self, tensor: Tensor) -> None:
        """A DataSource based around a created tensorflow Tensor.

        Good for testing purposes, but please use other sources for actual
        ML runs to prevent any data having to be hard-coded into the script.

        Args:
            tensor (Tensor): The Tensor to use as a source.
        """
        super().__init__()
        self.tensor = tensor

    def read(self) -> Tensor:
        """Returns the Tensor given as input before.

        Returns:
            Tensor: The Tensor used to initialise this object.
        """
        return self.tensor



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

Data = TypeVar("Data")


class DataSource(ABC, Generic[Data]):
    @abstractmethod
    def read(self) -> Data:
        """Read data from a given source.

        This method should be build ideally such that it provides consistent
        datasets every time it is called.

        Returns:
            Data: Data to be returned once read.
        """
        raise NotImplementedError



from setuptools import find_namespace_packages, setup

if __name__ == "__main__":
    version = "0.4.4"

    deps = [
        f"mlpype-base=={version}",
        "mlpype-sklearn",
        "xgboost>=1.6.2",
    ]
    strict_deps = [s.replace(">=", "==") for s in deps]

    setup(
        name="mlpype-xgboost",
        install_requires=deps,
        extras_require={"dev": strict_deps, "strict": strict_deps},
        packages=find_namespace_packages(include=["mlpype.*"]),
        version=version,
    )



from typing import Any, List, Type, Union

import tensorflow as tf
from pydantic import create_model
from tensorflow import DType, Tensor  # type: ignore

from mlpype.base.pipeline.type_checker import DataModel, TypeChecker


class TensorflowData(DataModel):
    data: list

    """An object that can be converted to a Tensorflow tensor."""

    def convert(self) -> Tensor:
        """Converts this object to a Tensorflow tensor.

        Returns:
            Tensor: The Tensorflow tensor contained by this object.
        """
        return tf.convert_to_tensor(self.data)  # type: ignore

    @classmethod
    def to_model(cls, data: Tensor) -> "TensorflowData":
        """Converts a Tensorflow tensor to a TensorflowData model, which can be serialised.

        Args:
            data (Tensor): A Tensorflow tensor to serialise.

        Returns:
            TensorflowData: A serialisable version of the tensor.
        """
        return TensorflowData(data=data.numpy().tolist())


class TensorflowTypeChecker(TypeChecker[Tensor]):
    def __init__(self) -> None:
        """Type checker for Tensorflow.

        Checks if the incoming data is of the correct non-first dimensions and dtype.
        """
        super().__init__()
        self.dims: tuple = tuple()
        self.dtype: Union[type, None] = None

    def fit(self, data: Tensor) -> "TensorflowTypeChecker":
        """Fit this Tensorflow TypeChecker to the given data.

        Args:
            data (Tensor): The data to fit.

        Returns:
            TensorflowTypeChecker: self.
        """
        self.dims = data.shape[1:]
        self.dtype = self._convert_dtype(data.dtype)
        return self

    def transform(self, data: Tensor) -> Tensor:
        """Checks if the given data fits the specifications this TypeChecker was fitted for.

        Args:
            data (Tensor):  The data to check.

        Returns:
            Tensor: data, if the data fits the specifications. Otherwise, an assertion error is thrown.
        """
        assert self.dtype is not None, "Please fit pipeline first"
        assert isinstance(data, Tensor), "Please provide a Tensorflow tensor!"
        assert (
            data.shape[1:] == self.dims
        ), f"Dimensions of Tensorflow tensors do not add up: {data.shape[1:]} vs {self.dims}"

        converted_type = self._convert_dtype(data.dtype)
        assert converted_type == self.dtype, f"Dtype of data does not add up: {converted_type} vs {self.dtype}"
        return data

    def _convert_dtype(self, dtype: DType) -> type:
        dtype_name = str(dtype)
        if "int" in dtype_name:
            return int
        elif "float" in dtype_name:
            return float
        elif "bool" in dtype_name:
            return bool
        else:
            return str

    def get_pydantic_type(self) -> Type[TensorflowData]:
        """Creates a Pydantic model for this data to handle serialisation/deserialisation.

        Returns:
            Type[TensorflowData]: A TensorflowData model that fits the data this wat fitted on.
        """
        base_iter: type = List[self.dtype]  # type: ignore

        for _ in range(len(self.dims)):
            base_iter = List[base_iter]  # type: ignore

        model = create_model("TensorflowData", data=(base_iter, ...), __base__=TensorflowData)

        return model

    @classmethod
    def supports_object(cls, obj: Any) -> bool:
        """Returns True if the object is a Tensorflow Tensor.

        Args:
            obj (Any): The object to check.

        Returns:
            bool: True if the given object is a Tensor, False otherwise.
        """
        return isinstance(obj, Tensor)



from typing import Any, Dict, List, Tuple, Type, Union

from pydantic import create_model
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql import SparkSession

from mlpype.base.pipeline.type_checker import DataModel, TypeChecker


class SparkData(DataModel):
    def convert(self) -> SparkDataFrame:
        """Converts this object to a spark DataFrame.

        Returns:
            SparkDataFrame: The spark DataFrame contained by this object.
        """
        list_format: List[Dict[str, Any]] = []
        length = None
        for name, value in self.__dict__.items():
            if length is None:
                length = len(value)
                list_format = [{} for _ in range(length)]
            else:
                assert length == len(value), f"Warning: {name} did not have the same length as others"
            for i, v in enumerate(value):
                list_format[i][name] = v

        spark_session = SparkSession.builder.getOrCreate()

        return spark_session.createDataFrame(list_format)

    @classmethod
    def to_model(cls, data: SparkDataFrame) -> "SparkData":
        """Converts a spark DataFrame to a SparkData model, which can be serialised.

        Args:
            data (SparkDataFrame): A spark DataFrame to serialise.

        Returns:
            SparkData: A serialisable version of the DataFrame.
        """
        rows = data.collect()
        columns: Dict[str, List[Any]] = {}
        for row in rows:
            for name, value in row.asDict().items():
                if name in columns:
                    columns[name].append(value)
                else:
                    columns[name] = [value]

        return cls(**columns)


class SparkTypeChecker(TypeChecker[SparkDataFrame]):
    def fit(self, data: SparkDataFrame) -> "SparkTypeChecker":
        """Fit this SparkTypeChecker to the given data.

        Args:
            data (SparkDataFrame): The data to fit.

        Returns:
            SparkTypeChecker: self
        """
        self.raw_types = self._convert_dtypes(dict(data.dtypes))
        return self

    def _convert_dtypes(self, dct: Dict[str, str]) -> Dict[str, Tuple[str, type]]:
        return {name: self._convert_dtype(string_type) for name, string_type in dct.items()}

    def _convert_dtype(self, string_type: str) -> Tuple[str, type]:
        conv_type: Union[Type, None] = None
        if string_type in ["str", "string"]:
            conv_type = str
        elif string_type in ["int", "bigint"]:
            conv_type = int
        elif string_type in ["float", "double"]:
            conv_type = float
        elif string_type == "bool":
            conv_type = bool
        else:
            raise ValueError(f"{string_type} not supported")

        return (string_type, conv_type)

    def transform(self, data: SparkDataFrame) -> SparkDataFrame:
        """Checks if the given data fits the specifications this TypeChecker was fitted for.

        Args:
            data (SparkDataFrame): The data to check.

        Returns:
            SparkDataFrame: data, if the data fits the specifications. Otherwise, an assertion error is thrown.
        """
        assert isinstance(data, SparkDataFrame), "Please provide a spark DataFrame!"
        colnames = list(self.raw_types.keys())
        for col in colnames:
            assert col in data.columns, f"`{col}` is missing from the dataset"
        dtype_dict = dict(data.dtypes)

        data = data.select(colnames)

        for name, (spark_type, _) in self.raw_types.items():
            actual_dtype = dtype_dict[name]
            assert (
                actual_dtype == spark_type
            ), f"Dtypes did not match up for col {name}: Expected {spark_type}, got {actual_dtype}"
        return data

    def get_pydantic_type(self) -> Type[SparkData]:
        """Creates a Pydantic model for this data to handle serialisation/deserialisation.

        Returns:
            Type[SparkData]: A SparkData model that fits the data this wat fitted on.
        """
        data_type = {
            name: (Union[List[dtype], Dict[str or int, dtype]], ...)  # type: ignore
            for name, (_, dtype) in self.raw_types.items()
        }

        model = create_model("SparkData", **data_type, __base__=SparkData)

        return model

    @classmethod
    def supports_object(cls, obj: Any) -> bool:
        """Returns True if the object is a Spark DataFrame.

        Args:
            obj (Any): The object to check.

        Returns:
            bool: True if the given object is a Spark DataFrame, False otherwise.
        """
        return isinstance(obj, SparkDataFrame)



from setuptools import find_namespace_packages, setup

if __name__ == "__main__":
    version = "0.4.4"

    deps = [
        f"mlpype-base=={version}",
        # We will provide absolute no guarantees that our integration will work with
        # EVERY version of pyspark. This has been developed under pyspark==3.2.1.
        "pyspark>=3.2.1",
    ]
    strict_deps = [s.replace(">=", "==") for s in deps]
    test_deps = ["pandas"]

    setup(
        name="mlpype-spark",
        install_requires=deps,
        extras_require={"dev": strict_deps + test_deps, "test": strict_deps + test_deps, "strict": strict_deps},
        packages=find_namespace_packages(include=["mlpype.*"]),
        version=version,
    )



from typing import Any, List, Type, Union

import numpy as np
from pydantic import create_model

from mlpype.base.pipeline.type_checker import DataModel, TypeChecker


class NumpyData(DataModel):
    data: list

    """An object that can be converted to a numpy array."""

    def convert(self) -> np.ndarray:
        """Converts this object to a numpy array.

        Returns:
            np.ndarray: The numpy array contained by this object.
        """
        return np.array(self.data)

    @classmethod
    def to_model(cls, data: np.ndarray) -> "NumpyData":
        """Converts a numpy array to a NumpyData model, which can be serialised.

        Args:
            data (np.ndarray): A numpy array to serialise.

        Returns:
            NumpyData: A serialisable version of the array.
        """
        return NumpyData(data=data.tolist())


class NumpyTypeChecker(TypeChecker[np.ndarray]):
    def __init__(self) -> None:
        """Type checker for numpy.

        Checks if the incoming data is of the correct non-first dimensions and dtype.
        """
        super().__init__()
        self.dims: tuple = tuple()
        self.dtype: Union[type, None] = None

    def fit(self, data: np.ndarray) -> "NumpyTypeChecker":
        """Fit this Numpy TypeChecker to the given data.

        Args:
            data (np.ndarray): The data to fit.

        Returns:
            NumpyTypeChecker: self.
        """
        self.dims = data.shape[1:]
        self.dtype = self._convert_dtype(data.dtype)
        return self

    def transform(self, data: np.ndarray) -> np.ndarray:
        """Checks if the given data fits the specifications this TypeChecker was fitted for.

        Args:
            data (np.ndarray):  The data to check.

        Returns:
            np.ndarray: data, if the data fits the specifications. Otherwise, an assertion error is thrown.
        """
        assert self.dtype is not None, "Please fit pipeline first"
        assert isinstance(data, np.ndarray), "Please provide a numpy array!"
        assert data.shape[1:] == self.dims, f"Dimensions of numpy arrays do not add up: {data.shape[1:]} vs {self.dims}"

        converted_type = self._convert_dtype(data.dtype)
        assert converted_type == self.dtype, f"Dtype of data does not add up: {converted_type} vs {self.dtype}"
        return data

    def _convert_dtype(self, dtype: np.dtype) -> type:
        dtype_name = dtype.name
        if "int" in dtype_name:
            return int
        elif "float" in dtype_name:
            return float
        elif "bool" in dtype_name:
            return bool
        else:
            return str

    def get_pydantic_type(self) -> Type[NumpyData]:
        """Creates a Pydantic model for this data to handle serialisation/deserialisation.

        Returns:
            Type[NumpyData]: A NumpyData model that fits the data this wat fitted on.
        """
        base_iter: type = List[self.dtype]  # type: ignore

        for _ in range(len(self.dims)):
            base_iter = List[base_iter]  # type: ignore

        model = create_model("NumpyData", data=(base_iter, ...), __base__=NumpyData)

        return model

    @classmethod
    def supports_object(cls, obj: Any) -> bool:
        """Returns True if the object is a numpy array.

        Args:
            obj (Any): The object to check.

        Returns:
            bool: True if the given object is a numpy array, False otherwise.
        """
        return isinstance(obj, np.ndarray)



from typing import Any, Dict, List, Optional, Type

from mlpype.base.data import DataSet
from mlpype.base.pipeline import Operator


class Pipe:
    def __init__(
        self,
        name: str,
        operator: Type[Operator],
        inputs: List[str],
        outputs: List[str],
        kw_args: Optional[Dict[str, Any]] = None,
        fit_inputs: Optional[List[str]] = None,
        skip_on_inference: bool = False,
    ) -> None:
        """A single step in a Pipeline.

        Used to fit an Operator to data and apply transformations to subsequent DataSets.

        Args:
            name (str): The name of the Pipe.
            operator (Type[Operator]): The Operator class that this Pipe should use.
            inputs (List[str]): A list of input dataset names used by this Pipe.
            outputs (List[str]): A list of output dataset names used by this Pipe.
            kw_args (Optional[Dict[str, Any]]): keyword arguments to initialise the Operator.
            fit_inputs: (Optional[List[str]]): optional additional arguments to fit().
                Will not be used in transform().
            skip_on_inference (Optional[bool]): Flag indicating if this step should be skipped
                at inference time. Useful to pre-process response variables in the pipeline.
        """
        assert "__" not in name, "Pipe names cannot contain the string `__`"
        if fit_inputs is None:
            fit_inputs = []
        if kw_args is None:
            kw_args = {}

        self.name = name
        self.operator_class = operator
        self.args = kw_args
        self.operator = operator(**kw_args)
        self.inputs = inputs
        self.outputs = outputs
        self.fit_inputs = fit_inputs
        self.skip_on_inference = skip_on_inference

    def fit(self, data: DataSet) -> "Pipe":
        """Fits the Pipe to the given DataSet.

        Args:
            data (DataSet): The DataSet to use in fitting.

        Returns:
            Pipe: This object.
        """
        self.operator.fit(*data.get_all(self.inputs), *data.get_all(self.fit_inputs))
        return self

    def transform(self, data: DataSet, is_inference: bool = False) -> DataSet:
        """Transforms the given data using this Pipe.

        This Pipe should be fitted first using fit().

        Args:
            data (DataSet): The DataSet to use in transforming.
            is_inference (Optional[bool]): Flag indicating if we're in inference
                mode for this transformation. We'll skip this step if
                skip_on_inference was set to True.

        Returns:
            DataSet: The transformed Data.
        """
        # skip this step if we're in inference mode and this Pipe is marked as such.
        if is_inference and self.skip_on_inference:
            return data

        transformed = self.operator.transform(*data.get_all(self.inputs))
        if len(self.outputs) < 2:
            transformed = [transformed]
        result = data.copy()
        result.set_all(self.outputs, transformed)
        return result

    def inverse_transform(self, data: DataSet, is_inference: bool = False) -> DataSet:
        """Inverse transforms the DataSet using this Pipe.

        Note that this is automatically done in reverse: the inverse steps
        will be applied from back to front.

        We do not inverse transform if not all outputs (the inputs for the reverse) are present.
        It is up to the user to make sure the inverse transformations
        work if used on a partial dataset (e.g. only the output data).
        In these cases, we return the `data` arg.

        Args:
            data (DataSet): The DataSet to use in inverse transforming.
            is_inference (Optional[bool]): Flag indicating if we're in inference
                mode for this inverse transformation. We'll skip this step if
                skip_on_inference was set to True.

        Returns:
            DataSet: The inverse transformed DataSet
        """
        # skip this step if we're in inference mode and this Pipe is marked as such.
        if is_inference and self.skip_on_inference:
            return data

        # We do not inverse transform if not all outputs (the inputs for the inverse) are present.
        # It is up to the user to make sure the inverse transformations
        # work if used on a partial dataset (e.g. only the output data).
        if not all([i in data for i in self.outputs]):
            return data
        inverse = self.operator.inverse_transform(*data.get_all(self.outputs))
        if len(self.inputs) < 2:
            inverse = [inverse]
        result = data.copy()
        result.set_all(self.inputs, inverse)
        return result

    def reinitialise(self, args: Dict[str, Any]) -> None:
        """Re-initialises this Pipe's Operator given a dict of arguments.

        Args:
            args (Dict[str, Any]): The dictionary of arguments to use in re-initialisation.
        """
        self.operator = self.operator_class(**args)

    def copy(self, args: Optional[Dict[str, Any]] = None) -> "Pipe":
        """Create a copy of this Pipe's Operator given a dict of arguments.

        Args:
            args (Optional[Dict[str, Any]]): The dictionary of arguments to use in re-initialisation.
                If set to None, we'll use the same arguments as before.

        Returns:
            A new copy of this Pipe, with a new Operator object.
        """
        if args is None:
            args = self.args

        return Pipe(
            self.name,
            self.operator_class,
            inputs=self.inputs,
            outputs=self.outputs,
            kw_args=args,
            fit_inputs=self.fit_inputs,
        )

    def __str__(self) -> str:
        """Create string representation of this Pipe.

        Returns:
            str: A string representation of this Pipe.
        """
        fit_section = f" (+ {self.fit_inputs})" if len(self.fit_inputs) > 0 else ""
        input_output_section = f"{self.inputs}{fit_section} -> {self.outputs}"

        return f"Pipe `{self.name}`, {input_output_section}: {self.operator}"



import json
from pathlib import Path
from typing import Any, List, Type, Union

import joblib
from pyspark.ml import Transformer

from mlpype.base.pipeline import Pipe, Pipeline
from mlpype.base.serialiser import Serialiser
from mlpype.spark.pipeline.spark_pipe import SparkPipe


class SparkSerialiser(Serialiser):
    """A Serialiser to integrate Spark with mlpype.

    It is highly recommended to initialise SparkSession manually before calling this,
    with all proper configuration done beforehand. A SparkSession is needed to
    serialise / deserialise objects stored using this class.
    """

    SUB_PIPE_PREFIX = "__sub_pipe_"
    BASE_PIPE_FILE = "pipe.pkl"
    SPARK_TRANSFORMER_FILE = "spark_transformer"
    SPARK_TRANSFORMER_CLASS_FILE = "spark_transformer_class"
    STEPS_FILE = "steps.json"

    def serialise(self, object: Any, file: Union[str, Path]) -> None:
        """Serialse a given object.

        If the object is:
            - A mlpype Pipeline: we serialise each step separately.
            - Otherwise, we use joblib.

        Args:
            object (Any): The object to serialise.
            file (Union[str, Path]): The path to serialise to.
        """
        if isinstance(object, Pipeline):
            self._serialise_pipeline(object, file)
        else:
            return self._serialise_joblib(object, file)

    def deserialise(self, file: Union[str, Path]) -> Any:
        """Deserialise the object in the given file.

        This function can handle mlpype Pipelines with PySpark elements.

        Args:
            file (Union[str, Path]): The file containing a python object to deserialise.

        Returns:
            Any: The deserialised object.
        """
        file = Path(file)
        if file.is_dir() and (file / self.STEPS_FILE).is_file():
            return self._deserialise_pipeline(file)
        else:
            return self._deserialise_joblib(file)

    def _serialise_pipeline(self, pipeline: Pipeline, file: Union[str, Path]) -> None:
        file = Path(file)

        file.mkdir(exist_ok=True)
        steps = []
        i = 0
        for step in pipeline:
            if isinstance(step, SparkPipe):
                name = step.name
                pipe_path = file / name
                pipe_path.mkdir(exist_ok=True)
                self._serialise_joblib(step, pipe_path / self.BASE_PIPE_FILE)

                # store the class for re-importing
                transformer: Transformer = step.fitted
                self._serialise_joblib(type(transformer), pipe_path / self.SPARK_TRANSFORMER_CLASS_FILE)
                transformer.save(str(pipe_path / self.SPARK_TRANSFORMER_FILE))
            elif isinstance(step, Pipe):
                name = step.name
                self._serialise_joblib(step, file / name)
            else:
                name = f"{self.SUB_PIPE_PREFIX}{i}"
                i += 1
                self._serialise_pipeline(step, file / name)

            steps.append(name)

        with open(file / self.STEPS_FILE, "w") as f:
            json.dump({"steps": steps}, f)

    def _serialise_joblib(self, object: Any, file: Union[str, Path]) -> None:
        """Serialise the given object to the given file.

        Args:
            object (Any): The object to serialise.
            file (Union[str, Path]): The file to serialise to.
        """
        joblib.dump(object, file)

    def _deserialise_pipeline(self, file: Union[str, Path]) -> Pipeline:
        file = Path(file)
        assert file.is_dir()

        with open(file / self.STEPS_FILE, "r") as f:
            steps: List[str] = json.load(f)["steps"]

        pipes = []
        for step in steps:
            step_path = file / step
            if step_path.is_file():
                pipes.append(self._deserialise_joblib(step_path))
            elif step_path.is_dir():
                if step.startswith(self.SUB_PIPE_PREFIX):
                    pipes.append(self._deserialise_pipeline(step_path))
                else:
                    pipe: SparkPipe = self._deserialise_joblib(step_path / self.BASE_PIPE_FILE)
                    fitted_class: Type[Transformer] = self._deserialise_joblib(
                        step_path / self.SPARK_TRANSFORMER_CLASS_FILE
                    )
                    pipe.fitted = fitted_class.load(str(step_path / self.SPARK_TRANSFORMER_FILE))
                    pipes.append(pipe)
            else:
                raise ValueError(f"{step_path} was not an existing path")

        return Pipeline(pipes)

    def _deserialise_joblib(self, file: Union[str, Path]) -> Any:
        """Deserialise the object in the given file.

        Args:
            file (Union[str, Path]): The file to deserialise.

        Returns:
            Any: The python object stored in the file.
        """
        return joblib.load(file)



from typing import Any, Dict


def get_args_for_prefix(prefix: str, dct: Dict[str, Any]) -> Dict[str, Any]:
    """Get all keys starting with `prefix` and remove the prefix.

    Args:
        prefix (str): The prefix to search. In the resulting dictionary these prefixes will
            be removed from the keys.
        dct (Dict[str, Any]): The dictionary to search for the given prefix.

    Returns:
        Dict[str, Any]: dct, reduced by only including keys that included the prefix and where
            that prefix is removed from the original key.
    """
    return {key.replace(prefix, "", 1): value for key, value in dct.items() if key.startswith(prefix)}



import typing
from abc import ABC, abstractmethod
from argparse import ArgumentParser
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)

from keras import Model as KerasBaseModel
from keras.losses import Loss
from keras.metrics import Metric
from keras.models import load_model as keras_load_model
from keras.optimizers import Optimizer
from tensorflow import Tensor  # type: ignore
from tensorflow.data import Dataset  # type: ignore
from tensorflow.random import set_seed  # type: ignore

from mlpype.base.experiment.argument_parsing import add_args_to_parser_for_class
from mlpype.base.model import Model
from mlpype.base.serialiser.joblib_serialiser import JoblibSerialiser

T = TypeVar("T", bound=KerasBaseModel)


class KerasPypeModel(Model[Tensor], ABC, Generic[T]):
    KERAS_MODEL_FILE = "keras_model"
    LOSS_FILE = "loss_func.pkl"

    def __init__(
        self,
        inputs: List[str],
        outputs: List[str],
        loss: Union[Callable, Loss],
        optimizer_class: Type[Optimizer],
        epochs: int = 10,
        batch_size: int = 32,
        learning_rate: float = 0.001,
        metrics: Optional[List[Union[Metric, Dict[str, Metric]]]] = None,
        model: Optional[T] = None,
        seed: int = 1,
        **model_args: Any,
    ) -> None:
        """A Model to integrate Keras models with the mlpype framework.

        Args:
            inputs (List[str]): Input names from the DataSet to use.
            outputs (List[str]): Output names from the DataSet to use.
            loss (Union[Callable, Loss]): The loss for your Keras Model.
            optimizer_class (Type[Optimizer]): The optimization class to use.
            epochs (int, optional): The number of epochs to use for training. Defaults to 10.
            batch_size (int, optional): The batch size to use for training. Defaults to 32.
            learning_rate (float, optional): The learning rate to use for training. Defaults to 0.001.
            metrics (Optional[List[Union[Metric, Dict[str, Metric]]]]): Additional metrics to
                use while training th emodel. Defaults to no metrics.
            model (Optional[T]): The Keras Model to train. By Default we'll use excess arguments
                to initialise the default model.
            seed (int, optional): The seed to initialise tensorflow with. Defaults to 1.
        """
        super().__init__(inputs, outputs, seed)
        if metrics is None:
            metrics = []

        if model_args is None:
            model_args = {}

        if model is None:
            model = self._init_model(model_args)
        self.model = model
        self.optimizer = optimizer_class(learning_rate=learning_rate)
        self.loss = loss
        self.metrics = metrics

        self.epochs = epochs
        self.batch_size = batch_size

    @abstractmethod
    def _init_model(self, args: Dict[str, Any]) -> T:
        raise NotImplementedError

    @classmethod
    def _get_annotated_class(cls) -> Type[KerasBaseModel]:
        return typing.get_args(cls.__orig_bases__[0])[0]

    def set_seed(self) -> None:
        """Sets the RNG seed."""
        set_seed(self.seed)

    def _fit(self, *data: Tensor) -> None:
        dataset = Dataset.from_tensor_slices(data).batch(self.batch_size)

        self.model.compile(optimizer=self.optimizer, loss=self.loss, metrics=self.metrics)
        self.model.fit(dataset, epochs=self.epochs)

    def _transform(self, *data: Tensor) -> Union[Iterable[Tensor], Tensor]:
        return self.model(*data)

    def _save(self, folder: Path) -> None:
        serialiser = JoblibSerialiser()
        serialiser.serialise(self.loss, folder / self.LOSS_FILE)
        self.model.save(folder / self.KERAS_MODEL_FILE)

    @classmethod
    def _load(cls: Type["KerasPypeModel"], folder: Path, inputs: List[str], outputs: List[str]) -> "KerasPypeModel":
        model: KerasBaseModel = keras_load_model(folder / cls.KERAS_MODEL_FILE)

        serialiser = JoblibSerialiser()
        loss = serialiser.deserialise(folder / cls.LOSS_FILE)
        optimizer = model.optimizer
        return cls(inputs=inputs, outputs=outputs, model=model, seed=1, loss=loss, optimizer_class=optimizer.__class__)

    @classmethod
    def get_parameters(cls: Type["KerasPypeModel"], parser: ArgumentParser) -> None:
        """Get and add parameters to initialise this class.

        KerasPypeModel's will work by requiring 2 ways to instantiate a Model:
            - through `model`, which is a keras model.
            - through parameters, which will instantiate the model internally.

        Arguments are taken from both the Keras Model class and the mlpype Model class.

        Args:
            parser (ArgumentParser): The ArgumentParser to add arguments to.
        """
        super().get_parameters(parser)
        BaseModel = cls._get_annotated_class()

        add_args_to_parser_for_class(parser, BaseModel, "model", [KerasBaseModel], excluded_args=[])

        add_args_to_parser_for_class(
            parser,
            cls,
            "model",
            [Model],
            excluded_args=["seed", "inputs", "outputs", "model", "loss", "optimizer_class"],
        )



import importlib
import logging
import os
import sys
from contextlib import contextmanager
from pathlib import Path
from types import ModuleType
from typing import Generator, List, Optional, Set, Union


@contextmanager
def switch_directory(directory: Union[str, Path]) -> Generator:
    """Allows you to temporarily switch Python over to a different directory.

    This changes the following:
        - The current working directory will change to `target_workspace`

    These changes will be undone once the context manager ends.

    Args:
        target_workspace (Union[str, Path]): The directory to switch to.
    """
    # create backups of the workspace and path.
    target_directory = Path(directory)
    old_workspace = os.getcwd()
    os.chdir(target_directory)
    # Yield so the normal code in the managed context can run.
    yield

    os.chdir(old_workspace)


@contextmanager
def switch_workspace(
    target_workspace: Union[str, Path], extra_files: Optional[List[Union[str, Path]]] = None
) -> Generator:
    """Allows you to temporarily switch Python over to using a different working directory.

    This changes the following:
        - The current working directory will change to `target_workspace`
        - `target_workspace` is added to the system path
        - Any files provided in `extra_files` will be assumed to follow a relative path
            from `target_workspace` to python files that will be imported. Variables
            (incl. classes and functions) defined in those files will be imported into
            the __main__ module, and the files will be added to the path. We do not overwrite variables
            already existing in __main__.

    These changes will be undone once the context manager ends.

    Args:
        target_workspace (Union[str, Path]): The directory to switch to.
        extra_files (Optional[List[Union[str, Path]]]): Any additional files whose contents should
            be added to the current python program. Defaults to no extra files.
    """
    if extra_files is None:
        extra_files = []

    # create backups of the workspace and path.
    target_workspace = Path(target_workspace)
    old_workspace = os.getcwd()
    old_sys_path = sys.path.copy()

    # Add each file to  the path, and import variables specific to that file into globals.
    main_lib = importlib.import_module("__main__")
    main_to_remove = _create_temporary_workspace(target_workspace, extra_files, main_lib)
    # Yield so the normal code in the managed context can run.
    yield

    _reset_workspace(old_workspace, old_sys_path, main_lib, main_to_remove)


def _find_all_py_files(files: List[Union[str, Path]]) -> List[str]:
    """Finds all .py files in the given list of files/directories.

    Args:
        files (List[Union[str, Path]]): A list of files / directories for which
            we want to collect the python files.

    Returns:
        List[str]: The list of python files in the given directories and files.
    """
    result = []
    for file in files:
        file_p = Path(file)
        assert file_p.exists(), f"`{file}` not found!"
        if Path(file_p).is_file():
            if str(file).endswith(".py"):
                result.append(str(file))
        else:
            # directory
            for root, _, walked_files in os.walk(file):
                for walked_file in walked_files:
                    if walked_file.endswith(".py"):
                        result.append(os.path.join(root, walked_file))
    return result


def _reset_workspace(
    old_workspace: str, old_sys_path: List[str], main_lib: ModuleType, main_to_remove: Set[str]
) -> None:
    os.chdir(old_workspace)
    sys.path = old_sys_path
    for var in main_to_remove:
        del main_lib.__dict__[var]


def _create_temporary_workspace(
    target_workspace: Path, extra_files: List[Union[str, Path]], main_lib: ModuleType
) -> Set[str]:
    """Perform the actual workspace switching.

    Args:
        target_workspace (Path): The directory to switch to.
        extra_files (List[Union[str ,  Path]]): Any additional files whose contents should
            be added to the current python program. Empty lists are accepted.
        main_lib (ModuleType): The __main__ module.

    Returns:
        Set[str]: The set of variable names that should be removed from __main__ again
            after the context manager quits.
    """
    logger = logging.getLogger(__name__)
    main_to_remove = set()

    # change directory and path
    os.chdir(target_workspace)
    sys.path.append(str(target_workspace))

    extra_files_walked = _find_all_py_files(extra_files)
    for file in extra_files_walked:
        if not file.endswith(".py"):
            logger.warning(f"Not adding {file} to path since it is not a python file!")
            continue

        # add the python file to the path. This makes sure we can import variables from other files.
        relative_file_name = str(os.path.join(".", file))
        sys.path.append(relative_file_name)

        # import the file as a module.
        module_name = file.replace("/", ".").replace(".py", "")
        lib = importlib.import_module(module_name)

        # find all variables defined in this file, avoiding things like numpy/pandas imports.
        sub_vars = {
            name: value
            for name, value in lib.__dict__.items()
            if hasattr(value, "__module__") and value.__module__ == module_name
        }

        # add the variable to the main module. This makes sure we can import variables from the
        # same 'main' file that was used to run the experiment.
        # out of safety reasons we only add a variable if it doesn't exist in 'main' yet.
        for var_name, value in sub_vars.items():
            if var_name not in main_lib.__dict__:
                main_to_remove.add(var_name)
                main_lib.__dict__[var_name] = value
    return main_to_remove



import shutil
from abc import ABC, abstractmethod
from pathlib import Path
from types import TracebackType
from typing import Any, Dict, Optional, Type, Union

from mlpype.base.model.model import Model
from mlpype.base.serialiser import Serialiser


class ExperimentLogger(ABC):
    @abstractmethod
    def __enter__(self) -> None:
        """Start the experiment."""
        return

    @abstractmethod
    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        """Stop the experiment."""
        return

    def log_metrics(self, dataset_name: str, metrics: Dict[str, Union[str, float, int, str, bool]]) -> None:
        """Logs the metrics for a given dataset.

        Metrics will be stored by default as `<dataset_name>_<metric_name>`.

        Args:
            dataset_name (str): The name of the dataset. Will be prepended to any metric name.
            metrics (Dict[str, Union[str, float, int, str, bool]]): A dictionary of metric names and values.
        """
        metrics = {f"{dataset_name}_{name}": metric for name, metric in metrics.items()}
        self._log_metrics(metrics)

    @abstractmethod
    def _log_metrics(self, metrics: Dict[str, Union[str, float, int, str, bool]]) -> None:
        """Perform the actual logging of metrics.

        Args:
            metrics (Dict[str, Union[str, float, int, str, bool]]): A dictionary of metric names and values.
        """
        raise NotImplementedError

    @abstractmethod
    def log_parameters(self, parameters: Dict[str, Any]) -> None:
        """Logs the parameters for a given run.

        These will by default be passed on with prefixes such as `model__`.

        Args:
            parameters (Dict[str, Any]): The parameters of a given run. Ideally these
                parameters should be no more complicated than string, float, int, bool, or a
                list of these.
        """
        raise NotImplementedError

    def log_model(self, model: Model, folder: Union[str, Path]) -> None:
        """Logs a Model for a given experiment.

        This function will write the model to the given location as well.

        Args:
            model (Model): The Model to be logged.
            MODEL_FOLDER (Union[str, Path]): The file to log the Model to.
        """
        model.save(folder)
        self.log_file(folder)

    def log_artifact(self, file: Union[str, Path], serialiser: Serialiser, object: Any) -> None:
        """Logs an artifact/file as part of an experiment.

        Args:
            file (Union[str, Path]): The file to serialise to and subsequently log.
            serialiser (Serialiser): Serialiser used to serialise the given object.
            object (Any): The object to log. This object will be serialised to the given file
                using the serialiser and logged using `log_file`
        """
        serialiser.serialise(object, file)
        self.log_file(file)

    def log_local_file(self, file: Union[str, Path], output_target: Union[str, Path]) -> None:
        """Logs a given local file / directory as part of an experiment.

        First the file / directory is copied to the output directory, then
        log_file is called to make it part of the experiment.

        Args:
            file (Union[str, Path]): The file / directory to log.
            output_target (Union[str, Path]): The target location of the file / directory.
        """
        file = Path(file)
        output_target = Path(output_target)
        output_target.parent.mkdir(exist_ok=True, parents=True)

        if file.is_file():
            shutil.copy(file, output_target)
        else:
            if output_target.exists():
                shutil.rmtree(output_target)
            shutil.copytree(file, output_target)

        self.log_file(output_target)

    @abstractmethod
    def log_file(self, file: Union[str, Path]) -> None:
        """Logs a given file as part of an experiment.

        Args:
            file (Union[str, Path]): The file to log.
        """
        raise NotImplementedError



from typing import Dict, Iterable, List, TypeVar

Data = TypeVar("Data")


class DataSet(Dict[str, Data]):
    """A DataSet consisting of multiple name-Data pairs.

    The Data can be multiple types of objects:
    - pandas DataFrames (or of other languages)
    - spark DataFrames
    - strings, integers, etc.
    """

    def get_all(self, keys: List[str]) -> List[Data]:
        """Returns all data associated with the given keys, in order.

        Args:
            keys (List[str]): The keys of all data to return.

        Returns:
            List[Data]: All data associated with the given keys.
        """
        return [self[key] for key in keys]

    @classmethod
    def from_dict(cls, dct: Dict[str, Data]) -> "DataSet[Data]":
        """Creates a DataSet from a dictionary.

        Args:
            dct (Dict[str, Data]): The dictionary containing data.

        Returns:
            DataSet[Data]: A DataSet based on the given dict.
        """
        return cls(**dct)

    def set_all(self, keys: List[str], data: Iterable[Data]) -> None:
        """Set all data to the given keys, in order.

        This assumes the keys and data set are of the same length.

        Args:
            keys (List[str]): The keys of the given data.
            data (Iterable[Data]): The data to store.
        """
        for key, d in zip(keys, data):
            self[key] = d

    def copy(self) -> "DataSet[Data]":
        """Create a shallow copy of this DataSet.

        This calls the superclass copy() function, but with more fitting type hints.

        Returns:
            DataSet[Data]: A shallow copy of this DataSet.
        """
        result = super().copy()
        return DataSet[Data](**result)



from typing import Dict, List, Optional, Union

from pyspark.ml.evaluation import JavaEvaluator

from mlpype.base.data.dataset import DataSet
from mlpype.base.evaluate.base_evaluator import BaseEvaluator
from mlpype.base.pipeline import Pipeline
from mlpype.spark.model import SparkModel


class SparkEvaluator(BaseEvaluator):
    def __init__(
        self,
        evaluator: JavaEvaluator,
        metrics: List[str],
    ) -> None:
        """Used to evaluate Spark models in mlpype.

        Args:
            evaluator (JavaEvaluator): One of the Evaluators from pyspark, e.g.
                psypark.ml.evaluation.RegressionEvaluator
            metrics (List[str]): The list of metrics to retrieve using the Evaluator.
                See the documentation of the one you use to know which are available.
        """
        super().__init__()
        self.evaluator = evaluator
        self.metrics = metrics

    def evaluate(
        self, model: SparkModel, data: DataSet, pipeline: Optional[Pipeline] = None
    ) -> Dict[str, Union[str, float, int, str, bool]]:
        """Evaluate the given model on the given dataset.

        We assume the model does not need to be transformed anymore if pipeline is None.

        Args:
            model (SparkModel): The SparkModel to evaluate.
            data (DataSet): The Dataset to use to evaluate the model.
            pipeline (Optional[Pipeline]): If not None, this will be used to transform `data` first.

        Returns:
            Dict[str, Union[str, float, int, str, bool]]: A dictionary of metric_name-value pairs. The result
                of the evaluation.
        """
        if pipeline is not None:
            data = pipeline.transform(data)

        predictions = model.transform_for_evaluation(data).get_all(model.outputs)

        assert len(predictions) == 1, "Expect exactly 1 output from a SparkModel."

        result = {}
        for metric in self.metrics:
            value = self.evaluator.setMetricName(metric).evaluate(predictions[0])
            result[metric] = value
        return result

    def __str__(self, indents: int = 0) -> str:
        """Generates a string representation of this SparkEvaluator.

        Args:
            indents (int, optional): The number of tabs to preceed this string
                representation. Defaults to 0.

        Returns:
            str: A string representation of this SparkEvaluator.
        """
        tabs = "\t" * indents
        metrics = ", ".join(self.metrics)
        return f"{tabs}{type(self.evaluator).__name__}: {metrics}"



from typing import Any, Dict

from pyspark.ml.regression import LinearRegression

from mlpype.spark.model.spark_model import SparkModel


class LinearSparkModel(SparkModel[LinearRegression]):
    def _init_model(self, args: Dict[str, Any]) -> LinearRegression:
        return LinearRegression(**args)



from setuptools import find_namespace_packages, setup

setup(
    name="mlpype",
    packages=find_namespace_packages(include=["mlpype.*"]),
    python_requires=">=3.8",
    version="0.4.4",
    license="MIT",
    author="Jeroen van den Hoven",
    url="https://github.com/jeroenvdhoven/mlpype",
)



import logging
from pathlib import Path
from types import TracebackType
from typing import Any, Dict, Optional, Type, Union

from git import InvalidGitRepositoryError
from git.repo import Repo
from mlflow import create_experiment  # type: ignore
from mlflow import get_experiment_by_name  # type: ignore
from mlflow import log_artifact  # type: ignore
from mlflow import log_metrics  # type: ignore
from mlflow import log_params  # type: ignore
from mlflow import set_experiment  # type: ignore
from mlflow import set_tag  # type: ignore
from mlflow import set_tracking_uri  # type: ignore
from mlflow import start_run  # type: ignore

from mlpype.base.logger import ExperimentLogger


class MlflowLogger(ExperimentLogger):
    def __init__(self, name: str, uri: str, artifact_location: Optional[str] = None) -> None:
        """A logger using mlflow for mlpype.

        Args:
            name (str): The name of the experiment. For runs done on databricks, make sure you preface
                this with a workspace, like /Users/<your databricks user name>/<experiment name>.
            uri (str): The tracking uri. Should most likely start with `http://` or `databricks`.
                It can also equal `databricks`.
            artifact_location (Optional[str]): The artificat uri location. This is needed if you set up
                your own artifact location, to make sure files are copied over to any remote tracking server.
                This is not required when using hosted databricks. Defaults to None, like in
        """
        super().__init__()
        logger = logging.getLogger(__name__)
        self.name = name

        if not uri.startswith("http://") and not uri.startswith("databricks"):
            logger.warning(f"Most often `uri` has to start with `http://` or `databricks`. Got: {uri}")
        self.uri = uri
        self.artifact_location = artifact_location
        self.run = None

    def __enter__(self) -> None:
        """Setup the mlflow experiment and call start_run."""
        set_tracking_uri(self.uri)
        experiment = get_experiment_by_name(self.name)

        if experiment is None:
            # this is a new experiment: create it explicitly with the artifact uri
            # TODO: make sure this works on S3
            create_experiment(self.name, self.artifact_location)

        set_experiment(experiment_name=self.name)

        self.run = start_run().__enter__()
        self.log_branch()

    def log_branch(self) -> None:
        """Set the current branch as a tag on the experiment."""
        assert self.run is not None, "Please start the experiment first"
        repo = self._find_encapsulating_repo(Path("."))

        if repo is None:
            branch_name = "unknown_branch"
        else:
            branch_name = repo.active_branch.name

        set_tag("git_branch", branch_name)

    def _find_encapsulating_repo(self, directory: Path) -> Union[Repo, None]:
        """Finds the current repo we're in, if any.

        We'll recursively move up from `directory` until we find a valid
        git repo.

        Args:
            directory (Path): The directory to start looking for. Usually set
                to the current directory.

        Returns:
            Repo or None: None if no repo is found, otherwise a GitPython Repo.
        """
        repo = None
        directory = directory.absolute()
        while str(directory) != "/":
            try:
                repo = Repo(directory)
                break
            except InvalidGitRepositoryError:
                directory = directory.parent
        return repo

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> None:
        """Exits a run started by __enter__.

        Args:
            exc_type (Optional[Type[BaseException]]): Passed to ActiveRun.__exit__()
            exc_val (Optional[BaseException]): Passed to ActiveRun.__exit__()
            exc_tb (Optional[TracebackType]): Passed to ActiveRun.__exit__()
        """
        assert self.run is not None, "Run not started, cannot exit!"
        self.run.__exit__(exc_type, exc_val, exc_tb)

    def _log_metrics(self, metrics: Dict[str, Union[str, float, int, str, bool]]) -> None:
        """Perform the actual logging of metrics in mlflow.

        Args:
            metrics (Dict[str, Union[str, float, int, str, bool]]): A dictionary of metric names and values.
        """
        log_metrics(metrics)

    def log_parameters(self, parameters: Dict[str, Any]) -> None:
        """Logs the parameters for a given run.

        These will by default be passed on with prefixes such as `model__`.

        Args:
            parameters (Dict[str, Any]): The parameters of a given run. Ideally these
                parameters should be no more complicated than string, float, int, bool, or a
                list of these.
        """
        log_params(parameters)

    def log_file(self, file: Union[str, Path]) -> None:
        """Logs a given file as part of an experiment.

        This only works well if:
            - you work with a databricks-hosted mlflow experiment.
            - you have set a artifact uri and configured your server properly.
            - you have run mlflow ui/server at the same folder where mlruns is located.

        Args:
            file (Union[str, Path]): The file to log.
        """
        log_artifact(str(file))



from typing import Optional

from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql import SparkSession

from mlpype.base.data.data_source import DataSource


class SparkSqlSource(DataSource[SparkDataFrame]):
    def __init__(
        self,
        query: str,
        spark_session: Optional[SparkSession] = None,
    ) -> None:
        """Use Spark SQL as an input data source.

        Args:
            spark_session (SparkSession): The current SparkSession.
            query (str): The query to run to get the data.
        """
        super().__init__()
        self.query = query
        if spark_session is None:
            spark_session = SparkSession.getActiveSession()
        self.spark_session = spark_session

    def read(self) -> SparkDataFrame:
        """Execute the query stored in this object.

        Returns:
            SparkDataFrame: The DataFrame generated by the query.
        """
        return self.spark_session.sql(self.query)



import json
import typing
from abc import ABC, abstractmethod
from argparse import ArgumentParser
from pathlib import Path
from typing import Any, Dict, Generic, Iterable, List, Optional, Type, TypeVar, Union

from pyspark.ml import Model as BaseSparkModel
from pyspark.ml import Predictor
from pyspark.sql import DataFrame as SparkDataFrame

from mlpype.base.data import DataSet
from mlpype.base.experiment.argument_parsing import add_args_to_parser_for_class
from mlpype.base.model import Model
from mlpype.base.serialiser import JoblibSerialiser

T = TypeVar("T", bound=Predictor)


class SparkModel(Model[SparkDataFrame], ABC, Generic[T]):
    mlpype_MODEL_CONFIG = "config.json"
    SPARK_PREDICTOR_PATH = "predictor"
    SPARK_MODEL_PATH = "spark_model"
    SPARK_MODEL_CLASS_PATH = "spark_model_class"

    def __init__(
        self,
        inputs: List[str],
        outputs: List[str],
        output_col: Optional[str] = None,
        model: Optional[BaseSparkModel] = None,
        predictor: Optional[T] = None,
        seed: int = 1,
        **model_args: Any,
    ) -> None:
        """A mlpype-compliant framework for using Spark Models.

        Args:
            inputs (List[str]): The name of the input DataFrame. Should be a list of 1 string.
            outputs (List[str]): The name of the output DataFrame. Should be a list of 1 string, the same as `inputs`
            output_col (Optional[str]): The name of the column where the model will put the output.
                Defaults to None, which means we won't select any columns and instead return the full output
                of the model.
            predictor (Optional[Predictor]): The Spark Predictor. If not set, we try to instantiate it
                using `model_args`
            model (Optional[BaseSparkModel]): The Spark Model. Defaults to None. If set to None,
                this model can't be serialised or used for inference.
            seed (int, optional): Spark Seed. Currently ignored, unfortunately. Defaults to 1.
        """
        assert len(inputs) == 1, "SparkML only requires a single DataFrame as input and output, the same one."
        assert len(outputs) == 1, "SparkML only requires a single DataFrame as input and output, the same one."
        assert inputs == outputs, "SparkML only requires a single DataFrame as input and output, the same one."
        super().__init__(inputs, outputs, seed)
        if predictor is None:
            predictor = self._init_model(model_args)

        self.model = model
        self.predictor = predictor
        self.output_col = output_col

    @abstractmethod
    def _init_model(self, args: Dict[str, Any]) -> T:
        raise NotImplementedError

    @classmethod
    def _get_annotated_class(cls) -> Type[T]:
        return typing.get_args(cls.__orig_bases__[0])[0]

    def set_seed(self) -> None:
        """Sets the RNG seed."""
        # TODO: set seed properly in pyspark
        print("Currently, setting a seed is not yet supported.")

    def fit(self, data: DataSet) -> "Model":
        """Fits the Model to the given DataSet.

        The DataSet should contain all inputs. For Spark models, this means 1 DataFrame
        that contains a feature vector and the label.

        Args:
            data (DataSet): The DataSet to fit this Model on.
        """
        self._fit(*data.get_all(self.inputs))
        return self

    def _fit(self, *data: SparkDataFrame) -> None:
        assert len(data) == 1, f"SparkML needs a single DataFrame as input, got {len(data)}"
        self.model = self.predictor.fit(data[0])

    def _transform(
        self, *data: SparkDataFrame, reduce_columns_if_possible: bool = True
    ) -> Union[Iterable[SparkDataFrame], SparkDataFrame]:
        assert len(data) == 1, f"SparkML needs a single DataFrame as input, got {len(data)}"
        assert self.model is not None, "Please fit this model before transforming data."

        result = self.model.transform(data[0])
        if self.output_col is not None and reduce_columns_if_possible:
            result = result.select(self.output_col)
        return result

    def transform_for_evaluation(self, data: DataSet) -> DataSet:
        """Applies the SparkModel to the given DataSet in an evaluation way.

        The DataSet should contain all inputs. This will not reduce the output dataset
        to the predicted columns such that we can use it in evaluation steps.

        Args:
            data (DataSet): The DataSet to transform using this Model.

        Returns:
            DataSet: The outputs as a Dataset.
        """
        result = self._transform(*data.get_all(self.inputs), reduce_columns_if_possible=False)
        if len(self.outputs) == 1:
            result = [result]  # type: ignore

        # type check handled by above check.
        return DataSet.from_dict({name: data for name, data in zip(self.outputs, result)})  # type: ignore

    def _save(self, folder: Path) -> None:
        assert self.model is not None, "Please fit this model before transforming data."

        serialiser = JoblibSerialiser()
        config = {"output_col": self.output_col}
        with open(folder / self.mlpype_MODEL_CONFIG, "w") as f:
            json.dump(config, f)

        self.predictor.save(str(folder / self.SPARK_PREDICTOR_PATH))
        self.model.save(str(folder / self.SPARK_MODEL_PATH))
        serialiser.serialise(type(self.model), str(folder / self.SPARK_MODEL_CLASS_PATH))

    @classmethod
    def _load(cls: Type["SparkModel"], folder: Path, inputs: List[str], outputs: List[str]) -> "SparkModel":
        serialiser = JoblibSerialiser()

        with open(folder / cls.mlpype_MODEL_CONFIG, "r") as f:
            config = json.load(f)
        output_col = config["output_col"]

        predictor_class: Type[Predictor] = cls._get_annotated_class()
        model_class: Type[BaseSparkModel] = serialiser.deserialise(str(folder / cls.SPARK_MODEL_CLASS_PATH))

        predictor: Predictor = predictor_class.load(str(folder / cls.SPARK_PREDICTOR_PATH))
        model: BaseSparkModel = model_class.load(str(folder / cls.SPARK_MODEL_PATH))

        return cls(inputs=inputs, outputs=outputs, predictor=predictor, model=model, output_col=output_col, seed=1)

    @classmethod
    def get_parameters(cls: Type["SparkModel"], parser: ArgumentParser) -> None:
        """Get and add parameters to initialise this class.

        SparkModel's will work by requiring 2 ways to instantiate a Model:
            - through `model`, which is a spark model.
            - through parameters, which will instantiate the model internally.

        Args:
            parser (ArgumentParser): The ArgumentParser to add arguments to.
        """
        super().get_parameters(parser)
        BaseModel = cls._get_annotated_class()

        add_args_to_parser_for_class(
            parser, BaseModel, "model", [], excluded_args=["seed", "inputs", "outputs", "model"]
        )



from dataclasses import dataclass
from logging import Logger, getLogger
from pathlib import Path
from typing import Any, Dict, Optional, Type, Union

from fastapi.applications import FastAPI
from fastapi.background import BackgroundTasks

from mlpype.base.data import DataSink
from mlpype.base.data.dataset import DataSet
from mlpype.base.deploy.inference import Inferencer
from mlpype.base.pipeline.type_checker import DataModel, DataSetModel


@dataclass
class mlpypeApp:
    name: str
    folder: Union[str, Path]
    tracking_servers: Optional[Dict[str, DataSink]] = None

    def __post_init__(self):  # type: ignore
        """Makes sure the folder is an actual Path."""
        if isinstance(self.folder, str):
            self.folder = Path(self.folder)

    def create_app(self) -> FastAPI:
        """Creates a FastAPI app for a saved mlpype experiment.

        Returns:
            FastAPI: The FastAPI server that can be served.
        """
        app = FastAPI()
        inferencer = self._load_model()
        logger = getLogger(self.name)

        InputType: Type[DataSetModel] = inferencer.input_type_checker.get_pydantic_types()
        OutputType: Type[DataSetModel] = inferencer.output_type_checker.get_pydantic_types()
        self._verify_tracking_servers(InputType, OutputType, logger)

        @app.get("/")
        async def home_page() -> str:
            logger.info("Homepage request")
            return f"Welcome to the mlpype FastAPI app for {self.name}"

        @app.post("/predict")
        async def predict(inputs: InputType, background: BackgroundTasks) -> OutputType:  # type: ignore
            logger.info("Prediction request")
            converted: DataSet = inputs.convert()  # type: ignore
            prediction = inferencer.predict(converted)
            result = OutputType.to_model(prediction)

            self._handle_tracking(converted, prediction, logger, background)
            logger.info("Finishing prediction request")
            return result

        return app

    def _handle_tracking(self, inputs: DataSet, outputs: DataSet, logger: Logger, background: BackgroundTasks) -> None:
        if self.tracking_servers is not None:
            logger.info("Logging records request")
            for name, sink in self.tracking_servers.items():
                try:
                    # Try to find the name of the dataset in the tracking servers.
                    ds: Union[DataModel, None] = None
                    if name in outputs:
                        ds = outputs[name]
                    elif name in inputs:
                        ds = inputs[name]

                    if ds is None:
                        logger.warning(f"Not storing data to sink `{name}` since no data by that name was found.")
                    else:
                        logger.info(f"Storing data to sink `{name}`.")
                        background.add_task(write_in_background, sink=sink, data=ds)

                except Exception as e:
                    logger.error(f"Encountered error while sending data to {name}: {str(e)}")

    def _verify_tracking_servers(
        self,
        inputs: Type[DataSetModel],
        outputs: Type[DataSetModel],
        logger: Logger,
    ) -> None:
        # Since it could be possible someone wants to log an intermediate dataset,
        # we can't put a hard restriction on logging only input or output data.
        # We will only log the latest version of a dataset though.
        if self.tracking_servers is not None:
            for name in self.tracking_servers:
                if not (name in inputs.__fields__ or name in outputs.__fields__):
                    logger.warning(f"No dataset named `{name}` found in the fields of input or output DataSetModels")

    def _load_model(self) -> Inferencer:
        assert isinstance(self.folder, Path)
        return Inferencer.from_folder(self.folder)


def write_in_background(sink: DataSink, data: Any) -> None:
    """Background function to write data to a given DataSink.

    Args:
        sink (DataSink): The Sink to write to.
        data (Any): The Data to write.
    """
    sink.write(data)



import logging
from abc import ABC, abstractclassmethod, abstractmethod
from typing import Any, Dict, List, Optional, Tuple, Type, Union

from pydantic import BaseModel, create_model

from mlpype.base.data.data_source import Data
from mlpype.base.data.dataset import DataSet
from mlpype.base.pipeline.operator import Operator
from mlpype.base.pipeline.pipe import Pipe


class DataModel(BaseModel, ABC):
    @abstractmethod
    def convert(self) -> Any:
        """Convert the Model to actual data (e.g. numpy or pandas).

        Returns:
            Any: The converted data.
        """

    @abstractclassmethod
    def to_model(cls, data: Any) -> "DataModel":
        """Convert the Model to actual data (e.g. numpy or pandas).

        Returns:
            Any: The converted data.
        """


class DataSetModel(BaseModel):
    def convert(self) -> DataSet:
        """Converts this DataSetModel to a DataSet, converting all data at the same time.

        Returns:
            DataSet: A DataSet based on this DataSetModel.
        """
        return DataSet({name: value.convert() for name, value in self.__dict__.items()})

    @classmethod
    def to_model(cls, ds: DataSet) -> "DataSetModel":
        """Converts a DataSet to a DataSetModel, which can be serialised.

        We convert data in the DataSet using known DataModels.

        Args:
            ds (DataSet): A DataSet to serialise.

        Returns:
            DataSetModel: A serialisable version of the DataSet.
        """
        dct = {}
        fields = cls.__fields__
        for name, dataset in ds.items():
            dct[name] = fields[name].type_.to_model(dataset)
        return cls(**dct)


class TypeChecker(Operator[Data], ABC):
    def __init__(self) -> None:
        """Defines a type checker class that can be used to check the format of new data."""
        super().__init__()

    @abstractmethod
    def fit(self, data: Data) -> "Operator":
        """Fits this Type Checker to the given Data.

        Args:
            data (Data): The data to fit.

        Returns:
            Operator: self.
        """

    @abstractmethod
    def transform(self, data: Data) -> Data:
        """Checks if the new data conforms to the specifications of the fitted data.

        Args:
            data (Data): The data to check.

        Returns:
            Data: data, if the data conforms. Otherwise an AssertionError will be thrown.
        """

    @abstractmethod
    def get_pydantic_type(self) -> Type[DataModel]:
        """Returns a Pydantic type for data serialisation/deserialistion based on this pipe.

        Returns:
            Type[DataModel]: The data model for the data this was fitted on.
        """

    @classmethod
    @abstractmethod
    def supports_object(cls, obj: Any) -> bool:
        """Class method to determine if this TypeChecker supports the given object.

        Args:
            obj (Any): The object to check.

        Returns:
            bool: True if this TypeChecker supports the given object, False otherwise.
        """


class TypeCheckerPipe(Pipe):
    def __init__(self, name: str, input_names: List[str], type_checker_classes: List[Type[TypeChecker]]) -> None:
        """A pipe that fully checks an incoming DataSet for type consistency.

        Args:
            name (str): The name of the pipe.
            input_names (List[str]): The names of datasets to be checked by this type checker.
                We highly recommend to make sure this encompasses the datasets used by
                the model at inference time (or the predicted datasets for the output type checker).
            type_checker_classes (List[Type[TypeChecker]]): A list of TypeChecker classes.
                E.g. PandasTypeChecker for pandas data (requires mlpype.sklearn).
        """
        super().__init__(
            name,
            DataSetTypeChecker,
            input_names,
            [],
            {
                "input_names": input_names,
                "type_checker_classes": type_checker_classes,
            },
        )
        self.input_names = input_names

    def get_pydantic_types(self, names: Optional[List[str]] = None) -> Type[DataSetModel]:
        """Generate a Pydantic model for the given dataset names.

        Args:
            names (Optional[List[str]]): The type checker for which we want data models
                to be returned. Defaults to all input datasets.

        Returns:
            Type[DataSetModel]: A pydantic model to serialise/deserialise data for the given
                names.
        """
        if names is None:
            names = self.input_names
        assert isinstance(
            self.operator, DataSetTypeChecker
        ), f"Operators of TypeCheckerPipes should be DataSetTypeChecker, got: `{type(self.operator)}`"
        return self.operator.get_pydantic_types(names)


class DataSetTypeChecker(Operator[Data]):
    def __init__(self, input_names: List[str], type_checker_classes: List[Type[TypeChecker]]) -> None:
        """A TypeChecker that conforms to the Operator class.

        Mainly used by TypeCheckerPipe for easy type checking of incoming data.

        Args:
            input_names (List[str]): The names of datasets to be checked by this type checker.
                We highly recommend to make sure this encompasses the datasets used by
                the model at inference time (or the predicted datasets for the output type checker).
            type_checker_classes (List[Type[TypeChecker]]): A list of TypeChecker classes.
                E.g. PandasTypeChecker for pandas data (requires mlpype.sklearn).
        """
        super().__init__()
        self.input_names = input_names
        self.type_checker_classes = type_checker_classes
        self.type_checkers: Dict[str, TypeChecker] = {}

    def fit(self, *data: Data) -> "Operator":
        """Fits all type checkers to the given data.

        Returns:
            Operator: self.
        """
        self.type_checkers = {}
        for ds_name, dataset in zip(self.input_names, data):
            type_checker_class = self._get_type_checker(dataset)

            if type_checker_class is None:
                logger = logging.getLogger(__name__)
                logger.warning(f"{ds_name} has no supported type checker!")
            else:
                checker = type_checker_class()
                checker.fit(dataset)
                self.type_checkers[ds_name] = checker
        return self

    def transform(self, *data: Data) -> Tuple[Data, ...]:
        """Checks if the given data fits the fitted Type Checkers.

        Returns:
            Tuple[Data, ...]: data, if all data fits the Type Checkers.
        """
        for ds_name, dataset in zip(self.input_names, data):
            assert ds_name in self.type_checkers, f"{ds_name} does not have a type checker"

            checker = self.type_checkers[ds_name]
            checker.transform(dataset)

        return data

    def _get_type_checker(self, data: Data) -> Union[Type[TypeChecker], None]:
        for type_checker in self.type_checker_classes:
            if type_checker.supports_object(data):
                return type_checker
        return None

    def get_pydantic_types(self, names: Optional[List[str]] = None) -> Type[DataSetModel]:
        """Generates a DataModel for the given input datasets after fitting.

        Args:
            names (Optional[List[str]]): The type checker for which we want data models
                to be returned. Defaults to all input datasets.

        Returns:
            Type[DataSetModel]: A pydantic model to serialise/deserialise data for the given
                input_names.
        """
        if names is None:
            names = self.input_names

        pydantic_types = {
            name: (checker.get_pydantic_type(), ...) for name, checker in self.type_checkers.items() if name in names
        }
        return create_model("DataSetModel", **pydantic_types, __base__=DataSetModel)


# class TypeCheckerOptions:
#     """Please register any Type Checkers you want to use!
#
#     For the ones defined in your script, just us the @register wrapper.
#     For the ones defined in standalone code (e.g. libs), put this import high
#     in the priority list, e.g. top level __init__
#
#     Returns:
#         _type_: _description_
#     """
#     registry: Dict[str, Tuple[type, Type[TypeChecker]]] = {}
#
#     @classmethod
#     def get_type_checker(cls, type_: type) -> Type[TypeChecker] or None:
#         for ref_type, type_checker in cls.registry.values():
#             if ref_type == type_:
#                 return type_checker
#         return None
#
#     @classmethod
#     def register(cls, name: str, type_: type):
#         def inner_wrapper(wrapped_class):
#             if name in cls.registry:
#                 print(f'Class {name} already exists. Will replace it')
#             cls.registry[name] = (type_, wrapped_class)
#             return wrapped_class
#         return inner_wrapper



import importlib
from pathlib import Path
from typing import Any, Callable, Dict, Optional, TypeVar, Union

import yaml
from jinja2 import Template

from mlpype.base.data.data_source import DataSource
from mlpype.base.data.dataset import DataSet

Data = TypeVar("Data")


class DataCatalog(Dict[str, DataSource[Data]]):
    """A collection of DataSources that together form a DataSet when loaded."""

    def read(self) -> DataSet:
        """Read all DataSources and generate a DataSet.

        Names of DataSources are preserved when loading the data.

        Returns:
            DataSet: The DataSet constructed from the DataSources.
        """
        return DataSet.from_dict({name: data.read() for name, data in self.items()})

    def __str__(self, indents: int = 0) -> str:
        """Create string representation of this DataCatalog.

        Args:
            indents (int, optional): The number of preceding tabs. Defaults to 0.

        Returns:
            str: A string representation of this DataCatalog.
        """
        tab = "\t" * indents
        return "\n".join([f"{tab}{name}: {source}" for name, source in self.items()])

    def read_one(self, name: str) -> Data:
        """Read a particular dataset from this DataCatalog.

        Args:
            name (str): The name of the dataset to read.

        Returns:
            Data: The data returned by the DataSource tied to `name`.
        """
        return self[name].read()

    @classmethod
    def from_yaml(cls, path: Union[str, Path], parameters: Optional[Dict[str, Any]] = None) -> "DataCatalog":
        """Read a catalog in from a configuration YAML file.

        We expect the following format:
        ```
        <dataset name>:
            callable: <python path to class, method on a class, or function.
                Should produce a DataSource.
            args:
                <name of argument>: <value>
        ```

        Values can be a plain value (like string, bool, etc) or a complex object.
        These will follow the same format as the callable-args structure of the
        DataSource's, but don't need to be DataSource's. e.g.:

        ```
        dataframe:
            callable: mlpype.sklearn.data.DataFrameSource
            args:
                df:
                    callable: pandas.DataFrame
                    args:
                        data:
                            x:
                                - 1.0
                                - 2.0
                            y:
                                - "a"
                                - "b"
        ```

        To use a method on a class, use `path.to.class:function`. This will only work for
        static or class methods.

        For security reasons, please make sure you trust any YAML file you read using this
        method! It will lead to code execution based on the imports, which can be abused.

        Parameters from the optional dictionary will be used for Jinja templating of
        strings in any of the arguments. To ensure non-string values are properly parsed,
        make sure they aren't surrounded by quotation marks in the YAML.

        Args:
            path (Union[str, Path]): The path to the YAML file.
            parameters (Optional[Dict[str, Any]]): an optional set of parameters to be used to
                do jinja templating on the YAML file's strings.

        Returns:
            DataCatalog: A DataCatalog with DataSources based on the YAML file.
        """
        data_sources = {}
        if parameters is None:
            parameters = {}

        path = Path(path)
        with open(path, "r") as f:
            contents = f.read()

            parsed_contents = Template(contents).render(parameters)
            configuration = yaml.safe_load(parsed_contents)
        assert isinstance(configuration, dict), "Cannot process YAML as DataCatalog: should be a dictionary."

        for dataset_name, dataset_params in configuration.items():
            dataset = cls._parse_object(dataset_params)
            assert isinstance(
                dataset, DataSource
            ), "Please make sure objects in your DataCatalog only translate to DataSources."
            data_sources[dataset_name] = dataset
        return DataCatalog(**data_sources)

    @classmethod
    def _parse_object(cls, dct: Dict[str, Any]) -> Any:
        """Recursively parse a complex dictionary to create objects.

        Args:
            dct (Dict[str, Any]): A dictionary containing 2 fields:
                - callable: A python path to a class or static/class method on a class.
                    Used to initiate the object.
                - args: a dictionary of args to instantiate the object with or call the
                    method with. Can be another complex object.

        Returns:
            Any: The object returned by calling `callable`
        """
        assert cls._is_valid_parseable_object(
            dct
        ), "DataCatalog: any dictionary parsed should have a `callable` and `args` entry."

        callable_ = cls._load_class(dct["callable"])
        args = dct["args"]
        assert isinstance(args, dict), "Arguments to a parseable object should be a dict."

        # Parse arguments recursively
        parsed_args = {}
        for arg_name, arg_value in args.items():
            if cls._is_valid_parseable_object(arg_value):
                parsed_value = cls._parse_object(arg_value)
            else:
                parsed_value = arg_value

            parsed_args[arg_name] = parsed_value

        return callable_(**parsed_args)

    @staticmethod
    def _load_class(full_path: str) -> Callable:
        """Load a callable from a Python import path.

        Supported functionality includes:
        - Importing a class, e.g. pandas.DataFrame.
        - Importing a static / class method, e.g. ds_range_allocation.data.catalog.DataCatalog:from_yaml

        Args:
            full_path (str): The path leading to the class or function to import.

        Returns:
            Callable: The class or function described in the `full_path` variable.
        """
        # check if we need to import a method.
        method_split = full_path.split(":")
        assert len(method_split) <= 2, "We do not accept paths with more than 1 `:`"
        callable_path = method_split[0]

        # for importing we need to split out the last part of the string.
        split_path = callable_path.split(".")
        module_path = ".".join(split_path[:-1])
        class_path = split_path[-1]

        # Import the module and get the class.
        module = importlib.import_module(module_path)
        callable = getattr(module, class_path)

        # Return the method if that was requested, otherwise just return the class.
        if len(method_split) > 1:
            return getattr(callable, method_split[1])
        else:
            return callable

    @staticmethod
    def _is_valid_parseable_object(dct: Dict[str, Any]) -> bool:
        return isinstance(dct, dict) and len(dct) == 2 and "callable" in dct and "args" in dct



from typing import Any, Callable, Dict, Optional, Tuple, Union

from hyperopt.base import Trials
from hyperopt.fmin import fmin, space_eval
from hyperopt.mongoexp import MongoTrials

from mlpype.base.experiment import Experiment


def create_optimisation_function(
    experiment_template: Experiment,
    target_metric: Tuple[str, str],
    minimise_target: bool,
    seed: int,
) -> Callable[[dict], Union[float, int]]:
    """Create an optimisation function that hyperopt can use to find the best model.

    Args:
        experiment_template (Experiment): An example Experiment that will be run
            multiple times to find the best hyperparameters. We will create clean copies
            of this Experiment for every run.
        target_metric (Tuple[str, str]): The target metric to optimise. Should be of the form
            (dataset name, metric name). The dataset name needs to match one in the DataSet from the Experiment.
            The metric name should match a metric from the BaseEvaluator in the Experiment.
        minimise_target (bool): A boolean indicating if the target metric should be minimised or not.
        seed (int): The seed used for training.

    Returns:
        Callable[[dict], Union[float, int]]: The optimisation function to be used by hyperopt's fmin
    """

    def optimise(hyper_params: Dict[str, Any]) -> Union[float, int]:
        """Optimise a given Experiment at the give hyperparameters.

        Args:
            hyper_params (Dict[str, Any]): The hyperparameters to initialise
                a new Experiment with. Make sure these are properly prefixed with
                model__ and pipeline__<pipe name>__, as you would initialise a
                regular Experiment from arguments.

        Returns:
            Union[float, int]: The performance of this Experiment on the given parameters
                on one metric and dataset.
        """
        experiment = experiment_template.copy(hyper_params, seed=seed)
        metrics = experiment.run()
        dataset, metric = target_metric

        result = metrics[dataset][metric]
        assert isinstance(result, (float, int)), "The target metric should be a float or integer, please fix this."

        # Since we use fmin, we'll need to multiply our target with -1 if we want to maxisime our target instead.
        if minimise_target:
            return result
        else:
            return -result

    return optimise


def optimise_experiment(
    experiment_template: Experiment,
    search_space: Dict[str, Tuple[str, Callable]],
    target_metric: Tuple[str, str],
    max_evals: int,
    minimise_target: bool,
    trial_type: str = "normal",
    mongo_url: Optional[str] = None,
    mongo_exp_key: str = "experiment",
    trials: Optional[Trials] = None,
    training_seed: int = 1,
    **kwargs: Any,
) -> Tuple[Union[int, None], Union[float, None], Dict[str, Any], Trials]:
    """Optimise a mlpype Experiment given a search space using Hyperopt.

    We attempt to create a clean copy of Experiments between runs to preserve run independency.
    To help this, please ensure the input/output checkers can be refitted without issue, and
    the DataSources provide the same data even if called multiple times.

    Args:
        experiment_template (Experiment): An Experiment that will serve as the template for each run.
            We try to ensure a new, clean Experiment is created for each run. Please allow this by
            not sharing important, trained variables between Models or Operators of the same class.
            Each call to any DataSource provided should also return the same dataset.
        search_space (Dict[str, Tuple[str, Callable]]): A search space defined in the standard hyperopt way.
            We suggest providing this as a dictionary in the form:
                {<argument name>: (<name>, <hyperopt callable like hp.choice(...)>)}
            Make sure these are properly prefixed with model__ and pipeline__<pipe name>__, as you would
            initialise a regular Experiment from arguments.
        target_metric (Tuple[str, str]): The target metric to optimise. Should be of the form
            (dataset name, metric name). The dataset name needs to match one in the DataSet from the Experiment.
            The metric name should match a metric from the BaseEvaluator in the Experiment.
        max_evals (int): The maximum number of runs to perform for this hyperoptimisation search. Note that
            hyperopt will not create more runs if you provide a trials object with more trials than max_evals.
        minimise_target (bool): A boolean indicating if the target metric should be minimised or not.
        trial_type (str): If `trials` is not provided, we use this to create a new Trials object. There are currently 3
            options available:
                - 'normal': A regular Trials object, stored locally.
                - 'mongo': A MongoTrials object, storing trials in a mongodb. You will need to provide `mongo_url` and
                    `mongo_exp_key` to choose this type of trial.
                - 'spark': A SparkTrials object will be created. Currently not supported yet, WIP. Providing your own
                    SparkTrials object should still work.
        mongo_url (Optional[str]): The link to the mongodb that should be used.
            Only used if `trial_type` == 'mongo', and is required in that case.
        mongo_exp_key (str): The experiment name in the mongodb that should be used.
            Only used if `trial_type` == 'mongo'. We encourage you to set this to distinguish between runs.
            The default is "experiment"
        trials (Optional[Trials]): An existing Trials object with previous runs of the optimisation.
            Useful if you want to continue an old optimisation run without losing information.
            By default a new Trials object will be created.
        training_seed (int): Seed used to initialise a new Experiment (and run).
        **kwargs: Additional arguments to be provided to hyperopt's `fmin`.

    Returns:
        A tuple consisting of:
            - The best trial index (or None if no good trial was found)
            - The performance of the best trial (or None if no good trial was found)
            - The arguments for the best trial.
            - The Trials object created by this run.
    """
    # TODO: SparkTrials setup integration.
    optimise = create_optimisation_function(experiment_template, target_metric, minimise_target, seed=training_seed)

    if not minimise_target:
        print("Heads up: printed targets will be maximised. In print output, they will be multiplied with -1.")

    # initialise a trials object, if needed.
    if trials is None:
        if trial_type == "mongo":
            assert mongo_url is not None, "mongo_url can not be None if you want to run hyperopt with Mongo."
            assert mongo_url.endswith("/jobs"), "hyperopt currently required the DB name to end in /jobs."
            trials = MongoTrials(mongo_url, exp_key=mongo_exp_key)
            mongo_url_sub = "://".join(mongo_url.split("://")[1:])
            print(
                f"""
    Please make sure you start hyperopt mongo workers in a separate terminal!
    Currently, you can use `hyperopt-mongo-worker --mongo={mongo_url_sub} --poll-interval=0.1` If this fails,
    check https://github.com/hyperopt/hyperopt/wiki/Parallelizing-Evaluations-During-Search-via-MongoDB"""
            )
        elif trial_type == "spark":
            raise ValueError(
                "Spark trials have not been implemented yet with auto-setup."
                + "Provide a SparkTrials object to the trials argument instead."
            )
        else:
            trials = Trials()

    # Run the optimisation
    opt_result = fmin(optimise, search_space, max_evals=max_evals, trials=trials, **kwargs)
    best_arguments = space_eval(search_space, opt_result)

    # Get the best trial ID and corresponding performance.
    if trials.best_trial is None:
        best_trial_id = None
        performance = None
    else:
        best_trial_id = trials.best_trial["tid"]
        performance = trials.best_trial["result"]["loss"]
        if not minimise_target:
            performance = -performance

    return best_trial_id, performance, best_arguments, trials



import json
from pathlib import Path
from typing import Optional, Type, Union

from mlpype.base.constants import Constants
from mlpype.base.data import DataCatalog, DataSet
from mlpype.base.experiment.experiment import Experiment
from mlpype.base.model import Model
from mlpype.base.pipeline import Pipeline
from mlpype.base.pipeline.type_checker import TypeCheckerPipe
from mlpype.base.serialiser.joblib_serialiser import JoblibSerialiser
from mlpype.base.serialiser.serialiser import Serialiser
from mlpype.base.utils.workspace import switch_workspace


class Inferencer:
    def __init__(
        self,
        model: Model,
        pipeline: Pipeline,
        input_type_checker: TypeCheckerPipe,
        output_type_checker: TypeCheckerPipe,
    ):
        """Provides a standard way of inferencing with mlpype models.

        Args:
            model (Model): The Model to use in inference.
            pipeline (Pipeline): The Pipeline to use in inference.
            input_type_checker (TypeCheckerPipe): The type checker used to determine incoming types.
            output_type_checker (TypeCheckerPipe): The type checker used to determine outgoing types.
        """
        self.model = model
        self.pipeline = pipeline
        self.input_type_checker = input_type_checker
        self.output_type_checker = output_type_checker

    def predict(self, data: Union[DataSet, DataCatalog]) -> DataSet:
        """Predicts using the given data using the Pipeline and Model.

        Args:
            data (Union[DataSet, DataCatalog]): The data to predict for.

        Returns:
            DataSet: The predictions from Model.
        """
        # TODO: inverse transformation after prediction
        if isinstance(data, DataCatalog):
            data = data.read()
        self.input_type_checker.transform(data)
        transformed = self.pipeline.transform(data, is_inference=True)
        predicted = self.model.transform(transformed)
        self.output_type_checker.transform(predicted)
        return predicted

    @classmethod
    def from_folder(cls: Type["Inferencer"], folder: Path, serialiser: Optional[Serialiser] = None) -> "Inferencer":
        """Loads a Inferencer from the results of an Experiment.

        We use the absolute version of the path to try and prevent loading issues.

        Args:
            folder (Path): The output folder from an Experiment, from which we load
                all required elements to make a inference pipeline.
            serialiser (Optional[Serialiser]): The Serialiser used to deserialise the pipeline and input/output
                type checkers. Defaults to a JoblibSerialiser.

        Returns:
            Inferencer: The inference pipeline that can predict for new data.
        """
        folder = folder.absolute()
        with open(folder / Constants.EXTRA_FILES, "r") as f:
            extra_files = json.load(f)["paths"]

        with switch_workspace(folder, extra_files):
            if serialiser is None:
                tmp_serialiser = JoblibSerialiser()
                serialiser = tmp_serialiser.deserialise(folder / Constants.SERIALISER_FILE)
            assert isinstance(serialiser, Serialiser), "Please provide a Serialiser!"

            model = Model.load(folder / Constants.MODEL_FOLDER)
            pipeline = serialiser.deserialise(folder / Constants.PIPELINE_FILE)
            input_type_checker = serialiser.deserialise(folder / Constants.INPUT_TYPE_CHECKER_FILE)
            output_type_checker = serialiser.deserialise(folder / Constants.OUTPUT_TYPE_CHECKER_FILE)

        return cls(
            model=model,
            pipeline=pipeline,
            input_type_checker=input_type_checker,
            output_type_checker=output_type_checker,
        )

    @classmethod
    def from_experiment(cls, experiment: Experiment) -> "Inferencer":
        """Generates an Inferencer from an experiment.

        We currently cannot check if the experiment has been trained already.

        Args:
            experiment (Experiment): The experiment to convert to a inferencing
                pipeline.

        Returns:
            Inferencer: An inferencing pipeline obtained from the given experiment.
        """
        return cls(
            model=experiment.model,
            pipeline=experiment.pipeline,
            input_type_checker=experiment.input_type_checker,
            output_type_checker=experiment.output_type_checker,
        )

    def __str__(self) -> str:
        """Create string representation of this Inferencer.

        Returns:
            str: A string representation of this Inferencer.
        """
        pipeline_part = "pipeline:\n" + self.pipeline.__str__(indents=1)
        model_part = f"model:\n\t{str(self.model)}"

        inputs = f"inputs:\n\t{str(self.input_type_checker)}"
        outputs = f"outputs:\n\t{str(self.output_type_checker)}"

        return "\n\n".join([inputs, outputs, pipeline_part, model_part])
